import glob
import importlib
import json
import os
import re
import socket
import threading
import time

from typing import Optional

from django.conf import settings
from django.core.cache import cache
from django.db import connections

from django_project_base.caching import CacheCounter
from django_project_base.caching.cache_queue import CacheQueue

DEFAULT_MAX_LOG_FILE_SIZE = 10000000
MAX_DATA_LOGGING_FILE_SIZE = 3000000

MATCH_DETAIL_QUERIES = re.compile(
    r"(rest/\w+)/((?:[0-9a-f]{8}(?:-[0-9a-f]{4}){3}-[0-9a-f]{12})|" r"(?:(?:[0-9a-f]{2}:){5}[0-9a-f]{2})|\d+)(/.*)?"
)


class ProfileRequest(object):
    response: object
    _start_time: tuple
    _end_time: tuple
    _process_function: callable
    _process_function_args: tuple
    _process_function_kwargs: dict
    _settings: dict
    _profile_path: str

    def __init__(
        self,
        settings: dict,
        process_function: callable,
        process_function_args: tuple,
        process_function_kwargs: dict,
    ):
        assert "REQUEST_METHOD" in settings
        assert "PATH_INFO" in settings
        settings.setdefault("QUERY_STRING", "")
        settings.setdefault("HTTP_HOST", socket.gethostname())
        self._settings = settings
        self._process_function = process_function
        self._process_function_args = process_function_args
        self._process_function_kwargs = process_function_kwargs

    def __enter__(self):
        # # Code to be executed for each request before
        # # the view (and later middleware) are called.
        tms = os.times()
        self._start_time = (int(time.time() * 1000), int(tms.user * 1000), int(tms.system * 1000))
        self._set_profiling_path(None, None)
        # # Get the response itself
        self.response = self._process_function(*self._process_function_args, **self._process_function_kwargs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # # Code to be executed for each request/response after
        # # the view is called.
        tms = os.times()
        self._end_time = (int(time.time() * 1000), int(tms.user * 1000), int(tms.system * 1000))
        self._do_profile(self.response, self._start_time, self._end_time)

    def _set_profiling_path(self, path_info, query_string):
        threading.current_thread().profiling_path = (path_info, query_string)

    def _get_profiling_path(self) -> tuple:
        return threading.current_thread().profiling_path

    def _get_path_info(self, path: str, params: Optional[str] = None):
        original_path = path
        path = path.strip("/")
        if "robots.txt" in path:
            return "other"
        if any(x in path for x in (".php", ".map")):
            return "other"

        if MATCH_DETAIL_QUERIES.match(path):
            return MATCH_DETAIL_QUERIES.sub("\\1\\3", path)

        if path.startswith("configure_site") and "&type=" in (params or ""):
            pos = params.find("&type=")
            return path + "/" + params[pos + 6 : pos + 9]

        if getattr(settings, "PROFILER_PATH_TRANSFORM", None):
            split_module_path: list = settings.PROFILER_PATH_TRANSFORM.split(".")
            module: str = ".".join(split_module_path[: (len(split_module_path) - 1)])
            method: str = split_module_path[-1]
            path = getattr(importlib.import_module(module), method)(
                path, {"original_path": original_path, "params": params}
            )

        return path

    def _get_queries(self, response):
        try:
            query_list = []
            for c in connections:
                con = connections[c]
                con.force_debug_cursor = True
                query_list.append(con.queries)
            _queries = filter(lambda x: x, query_list)
            return [q for sublist in _queries for q in sublist]
        except Exception as e:
            return ["exception getting queries: " + str(e)]

    def _do_profile(self, response, start_time, end_time):
        try:
            locs = locals()
            if "PATH_INFO" in self._settings:
                _profiling_path: tuple = next(iter(self._get_profiling_path()), None)
                path_info = _profiling_path or self._get_path_info(
                    str(self._settings["PATH_INFO"]), str(self._settings["QUERY_STRING"])
                )
                if path_info:
                    duration = (end_time[0] - start_time[0], end_time[1] - start_time[1], end_time[2] - start_time[2])
                    if hasattr(settings, "PROFILER_LONG_RUNNING_TASK_THRESHOLD"):
                        if duration[0] > settings.PROFILER_LONG_RUNNING_TASK_THRESHOLD:
                            queries = self._get_queries(response)
                            r_data = {
                                "timestamp": end_time[0] / 1000,
                                "duration": duration[0],
                                "queries": queries,
                                "PATH_INFO": path_info,
                            }
                            r_data.update(
                                {i: str(self._settings[i]) for i in ("HTTP_HOST", "REQUEST_METHOD", "QUERY_STRING")}
                            )
                            cache_ptr = CacheCounter("long_running_cmds_pointer", timeout=86400).incr(start=-1) % 50

                            cache.set("long_running_cmds_data%d" % cache_ptr, r_data, timeout=86400)

                        r_data = [path_info, duration[0], duration[1], duration[2]]
                        last_hour_running_cmds_key = f"last_hour_running_cmds{int(time.time()) // 10}"
                        last_hour_running_cmds_queue = CacheQueue.get_cache_queue(
                            last_hour_running_cmds_key, timeout=3600
                        )
                        last_hour_running_cmds_queue.rpush(json.dumps(r_data))

                    req_data = dict(
                        code=getattr(response, "status_code", None),
                        method=self._settings["REQUEST_METHOD"],
                        duration=duration[0],
                        timestamp=end_time[0] / 1000,
                        path_info=path_info,
                        pid=os.getpid(),
                        raw_path_info=self._settings.get("PATH_INFO", None),
                    )
                    files = glob.glob("/tmp/wsgi_performance.txt.*")
                    if not files:
                        with open("/tmp/wsgi_performance.txt.1", "a") as f:
                            f.write(json.dumps(req_data) + "\n")
                        files = ["/tmp/wsgi_performance.txt.1"]
                    file_num_array = list(map(lambda r: int(r.split(".")[-1]), files))
                    max_file = max(file_num_array)
                    min_file = min(file_num_array)
                    if max_file > 3:
                        if os.path.exists("/tmp/wsgi_performance.txt.1"):
                            os.remove("/tmp/wsgi_performance.txt.1")
                        if os.path.exists("/tmp/wsgi_performance.txt.2"):
                            os.remove("/tmp/wsgi_performance.txt.2")
                        os.rename("/tmp/wsgi_performance.txt.%d" % max_file, "/tmp/wsgi_performance.txt.1")
                        os.rename("/tmp/wsgi_performance.txt.%d" % (max_file - 1), "/tmp/wsgi_performance.txt.2")
                        for ff in filter(lambda f: f and int(f.split(".")[-1]) not in [1, 2], files):
                            if os.path.exists(ff):
                                os.remove(ff)
                        files = glob.glob("/tmp/wsgi_performance.txt.*")
                    if not files:
                        with open("/tmp/wsgi_performance.txt.1", "a") as f:
                            f.write(json.dumps(req_data) + "\n")
                    else:
                        if os.path.getsize("/tmp/wsgi_performance.txt.%d" % max_file) > MAX_DATA_LOGGING_FILE_SIZE:
                            with open("/tmp/wsgi_performance.txt.%d" % (max_file + 1), "a") as f:
                                f.write(json.dumps(req_data) + "\n")
                        else:
                            with open("/tmp/wsgi_performance.txt.%d" % max_file, "a") as f:
                                f.write(json.dumps(req_data) + "\n")
        except Exception as exc:
            try:
                error_file = "/tmp/wsgi_performance_error.txt"
                if os.path.exists(error_file) and os.path.getsize(error_file) > MAX_DATA_LOGGING_FILE_SIZE:
                    os.remove(error_file)
                with open(error_file, "a") as fe:
                    fe.write(str(exc) + "\n")
            except Exception:
                pass


def profile_middleware(get_response):
    # One-time configuration and initialization.

    def middleware(request):
        with ProfileRequest(request.META, get_response, (request,), {}) as pr:
            return pr.response

    return middleware



from typing import List, Optional

import requests

from django.core.cache import cache
from django.utils.translation import gettext_lazy as _
from requests import Timeout
from rest_framework import status

holidays_api_url: str = "https://date.nager.at/Api/v2/PublicHolidays/%d/%s"


class RetrieveHolidaysException(Exception):
    pass


def get_holidays(country_alpha_2_code: str, year: int) -> List[dict]:
    try:
        assert country_alpha_2_code and isinstance(country_alpha_2_code, str) and len(country_alpha_2_code) == 2, _(
            "Not valid country alpha 2 code"
        )
        assert isinstance(year, int) and year > 1900, _("Not a valid year")
        ck: str = f"country-holidays-{year}-{country_alpha_2_code}"
        cached_data: Optional[List[dict]] = cache.get(ck)
        if cached_data is not None:
            return cached_data
        response: requests.Response = requests.get(holidays_api_url % (year, country_alpha_2_code.upper()), timeout=6)
        assert response.status_code == status.HTTP_200_OK, _("Error retrieving data")
        holidays_data: List[dict] = response.json()
        cache.set(ck, holidays_data, timeout=None)
        return holidays_data
    except (AssertionError, Timeout) as e:
        raise RetrieveHolidaysException(_("Error retrieving holidays: {}".format(e))) from None



from django.test import TestCase

from django_project_base.notifications.base.phone_number_parser import PhoneNumberParser


class IsPhoneNumberValidTest(TestCase):
    def test_is_phone_number_valid(self):
        self.assertFalse(PhoneNumberParser.is_allowed("0903028"))
        self.assertFalse(PhoneNumberParser.is_allowed("1919"))
        self.assertTrue(PhoneNumberParser.is_allowed("03897234"))
        self.assertTrue(PhoneNumberParser.is_allowed("0038631697001"))
        self.assertTrue(PhoneNumberParser.is_allowed("+38676234567"))



import json

import swapper

from django.conf import settings
from django.contrib.sessions.middleware import SessionMiddleware as SessionMiddlewareBase
from django.utils.functional import SimpleLazyObject
from rest_framework.authentication import get_authorization_header


class ProjectNotSelectedError(NotImplementedError):
    def __init__(self, message: str, *args: object) -> None:
        super().__init__(message, *args)
        self.message = message


def selected_project_not_setup():
    raise ProjectNotSelectedError(
        """The functionality called requires settings variables that establish currently selected project.
        Check docs for DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES and its "project" setting.
        The setting needs to be declared and currently selected project passed at least once from the front-end
        """
    )


def load_selected_project(slug: str):
    def load():
        ProjectModel = swapper.load_model("django_project_base", "Project")
        try:
            return ProjectModel.objects.prefetch_related("owner").get(slug=slug)
        except ProjectModel.DoesNotExist:
            selected_project_not_setup()

    return load


class SessionMiddleware(SessionMiddlewareBase):
    def get_request_params(self, request):
        if request.method in ["POST", "PUT", "PATCH", "DELETE"]:
            body_data = json.loads(request.body.decode("utf-8"))
            if isinstance(body_data, list):
                return body_data[0]
            return body_data
        else:
            return getattr(request, request.method)

    def process_request(self, request):
        auth = get_authorization_header(request).split()
        if auth and auth[0].lower() == b"sessionid":
            session_key = auth[1].decode("utf-8")
            setattr(request, "_dont_enforce_csrf_checks", True)
        else:
            session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME)

        request.session = self.SessionStore(session_key)

        # determine if currently selected project has been passed with the request and set it to session if so
        # this requires UrlVarsMiddleware to have been installed before this middleware
        current_project_attr = (
            getattr(settings, "DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES", {})
            .get("project", {})
            .get("value_name", None)
        )

        # also set request.selected_project variables or throw errors on access if conditions not satisfied
        if current_project_attr:
            if curr_project_slug := getattr(request, current_project_attr, None):
                request.session[current_project_attr] = curr_project_slug
            else:
                curr_project_slug = request.session.get(current_project_attr, None)

            request.selected_project_slug = curr_project_slug
            request.selected_project = SimpleLazyObject(load_selected_project(curr_project_slug))
        else:
            request.selected_project_slug = SimpleLazyObject(selected_project_not_setup)
            request.selected_project = SimpleLazyObject(selected_project_not_setup)

    def process_response(self, request, response):
        if getattr(response, "returntype", None) == "json":
            process_response = super().process_response(request, response)
            process_response.cookies.pop("sessionid", None)
            process_response.cookies.pop("csrftoken", None)
            process_response.csrf_cookie_set = False

            return process_response

        return super().process_response(request, response)



from typing import Optional

import django
import swapper

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import connection, transaction
from django.db.models import ForeignKey
from django.utils.module_loading import import_string


class MergeUsersService:
    def handle(self, **kwargs):
        user = kwargs.get("user")
        if not user:
            return

        # TODO: handle multiple projects. In this case use signed in and has only on project available
        # if user logged in and hasnt got project selected we should run this script when project is selected
        project = swapper.load_model("django_project_base", "Project").objects.first()
        if not project:
            raise ValueError(f"Merge user service - no project found for user: {user.pk}")

        if group := self._find_group(user, project):
            try:
                with transaction.atomic():
                    self._change_user_pks(user, group)
                    if handler_function := getattr(settings, "MERGE_USERS_HANDLER", ""):
                        import_string(handler_function)(user=user, all_users=group.users, project=project)
                    group.delete()
            except Exception as e:
                import logging
                import traceback

                logger = logging.getLogger(__name__)
                logger.critical(e)
                logger.critical(str(traceback.format_exc()))

    def _find_group(self, user: "UserModel", project) -> Optional["MergeUserGroup"]:  # noqa:  F821
        MergeUserGroup = swapper.load_model("django_project_base", "MergeUserGroup")

        return next(
            iter(
                [
                    group
                    for group in MergeUserGroup.objects.filter(project=project)
                    if str(user.pk) in group.users.split(",")
                ]
            ),
            None,
        )

    def _change_user_pks(self, user, group):
        # TODO: this should be refactored when apps will support projects

        user_model = get_user_model()
        profile_model = swapper.load_model("django_project_base", "Profile")
        base_user_models = (user_model, profile_model, swapper.load_model("django_project_base", "MergeUserGroup"))
        users_to_merge = list(filter(lambda i: i and str(i) != str(user.pk), group.users.split(",")))
        project_model = swapper.load_model("django_project_base", "Project")

        db_tables = connection.introspection.table_names()

        for mdl in django.apps.apps.get_models(include_auto_created=True, include_swapped=True):
            if mdl not in base_user_models and not mdl._meta.abstract and not mdl._meta.swapped:
                if mdl._meta.db_table not in db_tables:
                    continue
                is_project_related = [
                    f for f in mdl._meta.fields if isinstance(f, ForeignKey) and f.related_model == project_model
                ]
                for fld in [
                    f
                    for f in mdl._meta.fields
                    if isinstance(f, ForeignKey) and (f.related_model in (user_model, profile_model))
                ]:
                    for usr in users_to_merge:
                        qs = mdl.objects.filter(**{fld.attname: fld.to_python(usr)})
                        if is_project_related:
                            p_fld = is_project_related[0]
                            qs = (
                                qs.filter(**{p_fld.attname: group.project.pk})
                                if getattr(group, "project", None)
                                else qs.filter()
                            )

                        m_items = []
                        if fld.many_to_one or fld.one_to_one:
                            m_items = mdl.objects.filter(**{fld.attname: fld.to_python(user.pk)})

                        for rec in qs:
                            try:
                                setattr(rec, fld.attname, fld.to_python(user.pk))
                                if (fld.many_to_one or fld.one_to_one) and m_items:
                                    try:
                                        rec.validate_unique()
                                        rec.save(update_fields=[fld.attname])
                                    except Exception:
                                        # record already exists, so delete record which would be duplicate
                                        rec.delete()
                                else:
                                    rec.save(update_fields=[fld.attname])
                            except Exception as e:
                                raise e
        for user in users_to_merge:
            profile = profile_model.objects.filter(pk=user_model._meta.pk.to_python(user)).first()
            if profile and profile.projects.all().count() > 1:
                # if user is on multiple projects skip deleting
                # TODO: this should be refactored when apps will support projects
                continue

            # handle all_auth tables
            if "account_emailaddress" in db_tables:
                with connection.cursor() as cursor:
                    cursor.execute("delete from public.account_emailaddress where user_id = %s", (user,))

            if "socialaccount_socialaccount" in db_tables:
                with connection.cursor() as cursor:
                    cursor.execute("truncate socialaccount_socialtoken;")
                    cursor.execute("delete from public.socialaccount_socialaccount where user_id = %s", (user,))

            profile_model.objects.filter(pk=user_model._meta.pk.to_python(user)).delete()
            user_model.objects.filter(pk=user_model._meta.pk.to_python(user)).delete()



import hashlib
import pickle

from typing import Any

from django.core.cache import cache
from django.db import models
from django.db.models import Model


class QuerySetWithCache(models.query.QuerySet):
    def cache_delete_pattern(self, pattern: str):
        if getattr(cache, "delete_pattern", None):
            cache.delete_pattern(pattern)
        else:
            # TODO: handle this !!!!!
            pass

    @property
    def cache_timeout(self) -> int:
        return 300

    @property
    def base_cache_key(self) -> str:
        return self.model.__name__.lower()

    def get_base_cache_key_item(self, pk: object) -> str:
        return "%s__pk__%s" % (self.base_cache_key, str(pk))

    def hash_args_kwargs(self, *args, **kwargs) -> str:
        return hashlib.md5(pickle.dumps((args, sorted(kwargs.items())))).hexdigest()

    def update(self, **kwargs):
        updated: Any = super().update(**kwargs)
        self.cache_delete_pattern("%s*" % self.base_cache_key)
        return updated

    def get(self, *args, **kwargs):
        ck: str = self.get_base_cache_key_item(self.hash_args_kwargs(args, kwargs))
        cached_item: Model = cache.get(ck)
        if cached_item:
            return cached_item
        item: Model = super().get(*args, **kwargs)
        cache.set(ck, item, timeout=self.cache_timeout)
        return item

    def create(self, **kwargs):
        item: Model = super().create(**kwargs)
        self.cache_delete_pattern("%s*" % self.base_cache_key)
        return item

    def filter(self, *args, **kwargs):
        return super().filter(*args, **kwargs)

    def list(self, *args, **kwargs):
        ck: str = "%s_%s_%s" % (self.base_cache_key, "filter", self.hash_args_kwargs(args, kwargs))
        cached_data: list = cache.get(ck)
        if cached_data is not None:
            return cached_data
        _data: list = list(super().filter(*args, **kwargs))
        cache.set(ck, _data, timeout=self.cache_timeout)
        return _data



import copy

from gettext import gettext
from typing import Union

import swapper

from django.conf import settings
from django.core.management import call_command
from django.db import transaction
from django.http import Http404
from django.shortcuts import get_object_or_404
from django.utils.module_loading import import_string
from django.utils.translation import gettext as _
from drf_spectacular.utils import extend_schema, OpenApiResponse
from dynamicforms import fields
from dynamicforms.action import TableAction, TablePosition
from dynamicforms.mixins import DisplayMode
from dynamicforms.serializers import DynamicModelMixin, DynamicModelSerializerMixin, ModelSerializer
from dynamicforms.template_render.layout import Column, Layout, Row
from dynamicforms.viewsets import ModelViewSet
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.permissions import IsAuthenticated, SAFE_METHODS
from rest_framework.request import Request
from rest_framework.response import Response

from django_project_base.account.middleware import ProjectNotSelectedError
from django_project_base.base.event import (
    ProjectSettingConfirmedEvent,
    ProjectSettingPendingResetEvent,
    SmsSenderChangedEvent,
)
from django_project_base.base.models import BaseProjectSettings
from django_project_base.base.permissions import CreateAny, IsProjectOwnerOrReadOnly
from django_project_base.constants import EMAIL_SENDER_ID_SETTING_NAME, SMS_SENDER_ID_SETTING_NAME
from django_project_base.utils import get_pk_name


class ProjectSerializer(DynamicModelSerializerMixin, ModelSerializer):
    template_context = dict(url_reverse="project-base-project")
    form_titles = {"table": _("Projects"), "new": _("New project"), "edit": _("Edit project")}
    MODEL_FUNC_SETTING_NAME = "DJANGO_PROJECT_BASE_PROJECT_MODEL_AT_RUNTIME"
    LAYOUT_FUNC_SETTING_NAME = "DJANGO_PROJECT_BASE_PROJECT_LAYOUT_AT_RUNTIME"

    def __init__(self, *args, is_filter: bool = False, **kwds):
        super().__init__(*args, is_filter=is_filter, **kwds)

        if self.context.get("view") and self.context["view"].format_kwarg == "componentdef":
            self.fields.fields["owner"].display_table = DisplayMode.SUPPRESS
            if self.context["view"].detail and self.instance.pk is None:
                # we are rendering new form
                self.fields.fields["owner"].display_form = DisplayMode.HIDDEN

    # logo = fields.FileField(display=DisplayMode.SUPPRESS, required=False)  # todo: not implemented UI
    owner = fields.AutoGeneratedField(display_table=DisplayMode.SUPPRESS, display_form=DisplayMode.HIDDEN)

    class Meta:
        model = swapper.load_model("django_project_base", "Project")
        exclude = ("logo",)  # TODO we currently don't support logos well. see DPB #3
        layout = Layout(Row("name"), Row("slug"), Row("description"))


class ProjectViewSet(DynamicModelMixin, ModelViewSet):
    serializer_class = ProjectSerializer
    permission_classes = (IsProjectOwnerOrReadOnly | CreateAny,)
    MODEL_FUNC_SETTING_NAME = "DJANGO_PROJECT_BASE_PROJECT_MODEL_AT_RUNTIME"

    def new_object(self: ModelViewSet):
        new_object = super().new_object()
        if self.request and self.request.user and self.request.user.is_authenticated:
            new_object.owner = getattr(
                self.request.user, swapper.load_model("django_project_base", "Profile")._meta.model_name
            )

        return new_object

    @staticmethod
    def _get_queryset_for_request(request):
        try:
            # had to copy this from the mixin because we don't have self here
            model_func = import_string(getattr(settings, ProjectViewSet.MODEL_FUNC_SETTING_NAME, None))
            model = model_func(None)
        except ImportError:
            model = swapper.load_model("django_project_base", "Project")

        qs = model.objects
        # todo: request.user.is_authenticated this should be solved with permission class
        if not request or not request.user or not request.user.is_authenticated:
            return qs.none()
        user_profile = getattr(request.user, swapper.load_model("django_project_base", "Profile")._meta.model_name)
        # projects where current user is owner
        owned_projects = qs.filter(owner=user_profile)
        # projects where user is member
        member_projects = qs.filter(members__member=user_profile)

        return (owned_projects | member_projects).distinct()

    def get_queryset(self):
        return ProjectViewSet._get_queryset_for_request(self.request)

    def get_serializer(self, *args, **kwargs):
        return super().get_serializer(*args, **kwargs)

    def get_permissions(self):
        if self.action == "get_current_project":
            return [IsAuthenticated()]
        else:
            return super().get_permissions()

    @extend_schema(
        description="Get currently selected project",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @action(
        methods=["GET"],
        detail=False,
        url_path="current",
        url_name="project-current",
    )
    def get_current_project(self, request: Request, **kwargs) -> Response:
        try:
            request.selected_project.get_deferred_fields()  # force immediate LazyObject evaluation
            serializer = self.get_serializer(request.selected_project)
        except ProjectNotSelectedError as e:
            raise NotFound(e.message)
        return Response(serializer.data)

    @extend_schema(
        description="Marks profile of calling user for deletion in future. Future date is determined " "by settings",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @get_current_project.mapping.put
    def update_current_profile(self, request: Request, **kwargs) -> Response:
        try:
            serializer = self.get_serializer(request.selected_project, data=request.data, many=False)
            serializer.is_valid(raise_exception=True)
            serializer.save()
        except ProjectNotSelectedError as e:
            raise NotFound(e.message)

        return Response(serializer.data)

    def get_object(self):
        SLUG_FIELD_NAME: str = settings.DJANGO_PROJECT_BASE_SLUG_FIELD_NAME

        lookup_field: str = self.lookup_field
        lookup_field_val: Union[str, int] = self.kwargs.get(self.lookup_field)

        proj_value_name = settings.DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES["project"]["value_name"]
        if getattr(self.request, proj_value_name, None) and lookup_field_val != "new":
            lookup_field_val = getattr(self.request, proj_value_name)

        def set_args(name: str) -> None:
            self.kwargs.pop(lookup_field, None)
            self.kwargs[name] = lookup_field_val
            self.lookup_field = name

        if lookup_field == "pk" or lookup_field == get_pk_name(self.get_queryset()):
            is_pk_auto_field: bool = self.get_queryset().model._meta.pk.get_internal_type() == "AutoField"
            try:
                int(lookup_field_val) if is_pk_auto_field and lookup_field_val else None
                return super().get_object()
            except (ValueError, Http404):
                set_args(SLUG_FIELD_NAME)
                return super().get_object()
        return super().get_object()

    def create(self, request, *args, **kwargs):
        create_response = super().create(request, *args, **kwargs)
        project = swapper.load_model("django_project_base", "Project").objects.get(slug=create_response.data["slug"])
        swapper.load_model("django_project_base", "ProjectMember").objects.create(project=project, member=request.user)
        project_settings_model = swapper.load_model("django_project_base", "ProjectSettings")
        project_settings_model.objects.create(
            name=EMAIL_SENDER_ID_SETTING_NAME,
            value=getattr(settings, "DEFAULT_EMAIL_SENDER_ID_SETTING_NAME", "") or gettext("Please enter value"),
            description=gettext("Email sender value for notifications"),
            value_type=BaseProjectSettings.VALUE_TYPE_CHAR,
            project=project,
            reserved=True,
        )
        project_settings_model.objects.create(
            name=SMS_SENDER_ID_SETTING_NAME,
            value=getattr(settings, "DEFAULT_SMS_SENDER_ID_SETTING_NAME", "") or gettext("Please enter value"),
            description=gettext("Sms sender value for notifications"),
            value_type=BaseProjectSettings.VALUE_TYPE_CHAR,
            project=project,
            reserved=True,
        )
        return create_response


class ProjectSettingsSerializer(ModelSerializer):
    template_context = dict(url_reverse="project-settings")

    def __init__(self, *args, is_filter: bool = False, **kwds):
        super().__init__(*args, is_filter=is_filter, **kwds)
        self.actions.actions = [a for a in self.actions.actions if a.name != "delete"]
        self.actions.actions.append(
            # TODO: https://taiga.velis.si/project/velis74-dynamic-forms/issue/837
            TableAction(
                position=TablePosition.FIELD_END,
                label="Reset pending",
                name="reset-pending",
                field_name="table_value",
                icon="refresh-outline",
                display_style=dict(
                    asButton=False,
                    showIcon=True,
                    showLabel=False,
                ),
            ),
        )
        request = self.context.get("request")
        if request and (request.user.is_superuser or request.user.is_staff):
            # TODO: https://taiga.velis.si/project/velis74-dynamic-forms/issue/837
            self.actions.actions.append(
                TableAction(
                    position=TablePosition.ROW_END,
                    label="Confirm active",
                    name="confirm-setting-active",
                ),
            )

    id = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)
    project = fields.PrimaryKeyRelatedField(
        display=DisplayMode.SUPPRESS, queryset=swapper.load_model("django_project_base", "Project").objects.all()
    )
    pending_value = fields.CharField(
        display=DisplayMode.SUPPRESS, required=False, allow_null=True, default=None, allow_blank=True
    )

    table_value = fields.CharField(
        source="value",
        display_table=DisplayMode.FULL,
        display_form=DisplayMode.SUPPRESS,
        label=gettext("Value"),
        read_only=True,
    )
    value = fields.CharField(
        display_table=DisplayMode.SUPPRESS,
        display_form=DisplayMode.FULL,
        required=True,
        allow_null=False,
        allow_blank=False,
    )

    def save(self, **kwargs):
        instance = copy.copy(self.instance)
        if instance:
            self.validated_data["pending_value"] = self.validated_data["value"]
            self.validated_data["value"] = instance.value

        if self.validated_data["value"] == self.validated_data["pending_value"]:
            self.validated_data["pending_value"] = None

        from django_project_base.base.event import EmailSenderChangedEvent

        EmailSenderChangedEvent(self.context["request"].user).trigger_changed(
            old_state=instance, new_state=self.Meta.model(**self.validated_data), payload=None
        )

        SmsSenderChangedEvent(self.context["request"].user).trigger_changed(
            old_state=instance, new_state=self.Meta.model(**self.validated_data), payload=None
        )
        saved = super().save(**kwargs)
        return saved

    def get_row_css_style(self, obj):
        if obj and obj.action_required:
            return "background-color:red;"
        return ""

    def to_representation(self, instance, row_data=None):
        representation = super().to_representation(instance, row_data)
        if instance and instance.pending_value is not None:
            representation["table_value"] = (
                f"{representation['value']} ({gettext('Pending')}: {instance.python_pending_value})"
            )
        return representation

    class Meta:
        model = swapper.load_model("django_project_base", "ProjectSettings")
        fields = (
            get_pk_name(model),
            "name",
            "table_value",
            "description",
            "value_type",
            "reserved",
            "value",
            "project",
            "pending_value",
        )
        layout = Layout(
            Row(Column("name")),
            Row(Column("value")),
            Row(Column("description")),
            Row(Column("value_type")),
            Row(Column("reserved")),
            size="large",
        )


class ProjectSettingsViewSet(ModelViewSet):
    permission_classes = [IsAuthenticated]

    def get_serializer_class(self):
        if (
            self.action == "retrieve" and self.detail is True and str(self.kwargs.get("pk", "")) != "new"
        ) or self.action in ("list", "update", "partial_update"):

            class AttrsReadOnlySer(ProjectSettingsSerializer):
                def __init__(self, *args, is_filter: bool = False, **kwds):
                    super().__init__(*args, is_filter=is_filter, **kwds)
                    if (
                        (view := kwds.get("context", {}).get("view"))
                        and view.detail is True
                        and str(view.kwargs.get("pk", "")) != "new"
                        and (sett := ProjectSettingsSerializer.Meta.model.objects.filter(pk=view.kwargs["pk"]).first())
                        and sett.reserved is True
                    ):
                        if view.action == "retrieve":
                            self.fields.fields["reserved"].read_only = True
                            self.fields.fields["name"].read_only = True
                            self.fields.fields["value_type"].read_only = True
                        else:
                            self.fields.fields["name"].required = False
                            self.fields.fields["value_type"].required = False

                def validate(self, attrs):
                    validated = super().validate(attrs)
                    if self.instance.reserved:
                        if "reserved" in validated and not validated["reserved"]:
                            raise PermissionDenied
                        if "name" in validated and validated["name"] != self.instance.name:
                            raise PermissionDenied
                        if "value_type" in validated and validated["value_type"] != self.instance.value_type:
                            raise PermissionDenied
                    return validated

                class Meta(ProjectSettingsSerializer.Meta):
                    pass

            return AttrsReadOnlySer

        return ProjectSettingsSerializer

    def initialize_request(self, request, *args, **kwargs):
        req = super().initialize_request(request, *args, **kwargs)
        if req.method.upper() not in SAFE_METHODS:
            try:
                req.data["project"] = request.selected_project.pk
            except ProjectNotSelectedError as e:
                raise NotFound(e.message)
        return req

    def get_queryset(self):
        try:
            return (
                self.get_serializer()
                .Meta.model.objects.filter(project__slug=self.request.selected_project.slug)
                .exclude(value_type=BaseProjectSettings.VALUE_TYPE_CUSTOM)
                .order_by("name")
            )
        except ProjectNotSelectedError:
            return self.get_serializer().Meta.model.objects.none()

    def handle_create_validation_exception(self, e, request, *args, **kwargs):
        if getattr(e, "model-validation", False):
            raise ValidationError({e.detail: e.default_code})
        return super().handle_create_validation_exception(e, request, *args, **kwargs)

    def list(self, request, *args, **kwargs):
        list_response = super().list(request, *args, **kwargs)
        pending_settings = (
            self.get_queryset()
            .filter(pending_value__isnull=False)
            .exclude(pending_value="")
            .order_by(f"-{get_pk_name(self.get_serializer_class().Meta.model)}")
        )
        if pending_settings.exists():
            pk_name = get_pk_name(self.get_serializer_class().Meta.model)
            list_response.set_cookie(
                "setting-verification",
                f"{'*'.join(list(map(str, pending_settings.values_list(pk_name, flat=True))))}",
                expires=24 * 60 * 60,
                samesite="Lax",
            )
            return list_response
        list_response.delete_cookie("setting-verification")
        return list_response

    def _get_pk_from_request(self, request: Request) -> int:
        model = self.get_serializer_class().Meta.model
        pk_name = get_pk_name(model)
        pk = request.data.get(pk_name)
        if not pk:
            raise ValidationError({pk_name: [gettext("required")]})
        return pk

    @extend_schema(exclude=True)
    @transaction.atomic()
    @action(
        detail=False,
        methods=["POST"],
        url_name="confirm-setting",
        url_path="confirm-setting",
    )
    def confirm_pending_setting(self, request) -> Response:
        ProjectSettingConfirmedEvent(user=request.user).trigger(
            payload=get_object_or_404(self.get_serializer_class().Meta.model, pk=self._get_pk_from_request(request))
        )
        return Response()

    @extend_schema(exclude=True)
    @transaction.atomic()
    @action(
        detail=False,
        methods=["POST"],
        url_name="reset-pending",
        url_path="reset-pending",
    )
    def reset_pending_setting(self, request) -> Response:
        setting = get_object_or_404(self.get_serializer_class().Meta.model, pk=self._get_pk_from_request(request))
        if setting.pending_value is not None:
            ProjectSettingPendingResetEvent(user=request.user).trigger(payload=setting)
        return Response()

    @extend_schema(exclude=True)
    @transaction.atomic()
    @action(
        detail=False,
        methods=["POST"],
        url_name="confirm-setting-active",
        url_path="confirm-setting-active",
    )
    def confirm_setting_active(self, request) -> Response:
        if not (request.user.is_superuser or request.user.is_staff):
            raise PermissionDenied
        setting = get_object_or_404(self.get_serializer_class().Meta.model, pk=self._get_pk_from_request(request))
        call_command(
            "confirm_setting",
            setting.project.pk,
            setting.name,
        )
        return Response()



from rest_framework.permissions import BasePermission, SAFE_METHODS

from django_project_base.account.middleware import ProjectNotSelectedError
from django_project_base.base.models import BaseProfile, BaseProject


class IsSuperUser(BasePermission):
    """
    Allows access only to admin users.
    """

    def has_permission(self, request, view):
        return bool(request.user and request.user.is_authenticated and request.user.is_superuser)


def can_user_hijack_another_user(hijacker, hijacked):
    # TODO: should this be solved by using roles
    return hijacker.is_authenticated and (hijacker.is_superuser or hijacker.is_staff)


def project_is_selected(project: BaseProject) -> bool:
    try:
        project.get_deferred_fields()  # force immediate LazyObject evaluation
    except ProjectNotSelectedError:
        return False
    return True


def is_authenticated(user: BaseProfile) -> bool:
    return bool(user and user.is_authenticated)


def is_project_member(user: BaseProfile, project: BaseProject) -> bool:
    return (
        is_authenticated(user)
        and project_is_selected(project)
        and (project.members.filter(member=user).exists() or is_project_owner(user, project))
    )


def is_project_owner(user: BaseProfile, project: BaseProject) -> bool:
    return is_authenticated(user) and project_is_selected(project) and user == project.owner


def is_superuser(user: BaseProfile) -> bool:
    return is_authenticated(user) and user.is_superuser


class IsProjectOwner(BasePermission):
    """
    Allows access only to project owners.
    """

    def has_permission(self, request, view):
        return is_superuser(request.user) or is_project_owner(request.user, request.selected_project)


class IsProjectMember(BasePermission):
    """
    Allows access only to project owners.
    """

    def has_permission(self, request, view):
        return is_superuser(request.user) or is_project_member(request.user, request.selected_project)


class IsProjectOwnerOrMemberReadOnly(BasePermission):
    """
    Allows access only to project owners.
    """

    def has_permission(self, request, view):
        from django_project_base.account.rest.project_profiles import ProjectProfilesViewSet

        if isinstance(view, ProjectProfilesViewSet):
            # this is a special case for user accounts: any user may write to their own account from anywhere
            # so if the user is trying to access their own account, we allow it, otherwise we run the default check
            try:
                instance = view.get_object()
                if instance == request.user:
                    return True
            except AssertionError:
                # will except in view.get_object() when the method is not PUT, PATCH, RETRIEVE
                pass

        return (
            is_superuser(request.user)
            or is_project_owner(request.user, request.selected_project)
            or ((request.method in SAFE_METHODS) and is_project_member(request.user, request.selected_project))
        )


class CreateAny(BasePermission):
    """
    Allows POST / create to anyone
    """

    def has_permission(self, request, view):
        return request.method == "POST"


class IsProjectOwnerOrReadOnly(BasePermission):
    """
    Allows access only to project owners.
    """

    def has_permission(self, request, view):
        return (
            is_superuser(request.user)
            or is_project_owner(request.user, request.selected_project)
            or (request.method in SAFE_METHODS)
        )


class IsProjectMemberOrAuthenticatedReadOnly(BasePermission):
    """
    Allows access only to project owners.
    """

    def has_permission(self, request, view):
        return is_project_member(request.user, request.selected_project) or (
            is_authenticated(request.user) and (request.method in SAFE_METHODS)
        )



import json
import logging
import uuid

from typing import Optional

from django.utils.module_loading import import_string
from rest_framework import viewsets
from rest_framework.authentication import BasicAuthentication, TokenAuthentication
from rest_framework.exceptions import ErrorDetail, ValidationError
from rest_framework.permissions import IsAuthenticated, SAFE_METHODS
from rest_framework.response import Response

from django_project_base.base.exceptions import LicenseActionNotImplementedException
from django_project_base.notifications.base.channels.integrations.provider_integration import ProviderIntegration
from django_project_base.notifications.models import DeliveryReport


class DeliveryReportViewSet(viewsets.ViewSet):
    permission_classes = [IsAuthenticated]

    authentication_classes = [TokenAuthentication, BasicAuthentication]

    _primary_key_names = ["id", "pk", "guid", "uuid"]

    def __find_dlr_id(self, params: dict) -> Optional[str]:
        id = next(filter(lambda i: i.lower() in self._primary_key_names, params.keys()), None)
        if id and params.get(id):
            uid = str(params[id])
            try:
                uuid.UUID(uid)
            except ValueError:
                raise ValidationError(ErrorDetail(string="Invalid identifier", code="invalid"))
            return uid
        logger = logging.getLogger("django")
        logger.exception(
            f"Id not found for delivery report {self.request.query_params} {self.request.data} {self.request.user}"
        )
        return None

    def __update_dlr(self, params) -> Optional[DeliveryReport]:
        if (id := self.__find_dlr_id(params)) and (dlr := DeliveryReport.objects.filter(pk=id).first()):
            dlr.payload = json.dumps(params)
            dlr.save(update_fields=["payload"])
            return dlr
        return None

    def __parse_delivery_report_payload(self, dlr: DeliveryReport):
        provider: ProviderIntegration = import_string(dlr.provider)()
        provider.parse_delivery_report(dlr)

    def list(self, request) -> Response:
        return self._handle_request(request)

    def create(self, request) -> Response:
        return self._handle_request(request)

    def _handle_request(self, request) -> Response:
        try:
            payload = request.data if request.method.upper() not in SAFE_METHODS else request.query_params
            if dlr := self.__update_dlr(payload):
                self.__parse_delivery_report_payload(dlr)
                return Response({0: "NO_ERROR"})
            return Response({400: "Invalid"})
        except Exception as e:
            logger = logging.getLogger("django")
            logger.exception(e)
            return Response({400: "Invalid"})

    def retrieve(self, request, pk=None):
        raise LicenseActionNotImplementedException()

    def update(self, request, pk=None):
        raise LicenseActionNotImplementedException()

    def partial_update(self, request, pk=None):
        raise LicenseActionNotImplementedException()

    def destroy(self, request, pk=None):
        raise LicenseActionNotImplementedException()



from django.apps import AppConfig

from django_project_base.notifications import NOTIFICATIONS_APP_ID
from django_project_base.notifications.settings import NOTIFICATIONS_SETTINGS
from django_project_base.settings_parser import parse_settings


class DjangoProjectBaseNotifyConfig(AppConfig):
    name = NOTIFICATIONS_APP_ID
    verbose_name = "Django Project Base Notifications"

    def ready(self):
        parse_settings(NOTIFICATIONS_SETTINGS)



import json

from math import ceil
from typing import Optional, Union

import requests
import swapper

from django.conf import Settings
from django.contrib.auth import get_user_model
from requests.auth import HTTPBasicAuth
from rest_framework.status import is_success

from django_project_base.celery.settings import NOTIFICATION_QUEABLE_HARD_TIME_LIMIT
from django_project_base.notifications.base.channels.channel import Recipient
from django_project_base.notifications.base.channels.integrations.provider_integration import ProviderIntegration
from django_project_base.notifications.models import DeliveryReport, DjangoProjectBaseNotification

# -*- coding: utf8 -*-
"""
Created on Jul 10, 2016
@author: Dayo
"""


class SMSCounter(object):
    # https: // raw.githubusercontent.com / dedayoa / sms - counter - python / master / sms_counter / main.py
    GSM_7BIT = "GSM_7BIT"
    GSM_7BIT_EX = "GSM_7BIT_EX"
    UTF16 = "UTF16"
    GSM_7BIT_LEN = GSM_7BIT_EX_LEN = 160
    UTF16_LEN = 70
    GSM_7BIT_LEN_MULTIPART = GSM_7BIT_EX_LEN_MULTIPART = 153
    UTF16_LEN_MULTIPART = 67

    @classmethod
    def _get_gsm_7bit_map(cls):
        gsm_7bit_map = [
            10,
            12,
            13,
            32,
            33,
            34,
            35,
            36,
            37,
            38,
            39,
            40,
            41,
            42,
            43,
            44,
            45,
            46,
            47,
            48,
            49,
            50,
            51,
            52,
            53,
            54,
            55,
            56,
            57,
            58,
            59,
            60,
            61,
            62,
            63,
            64,
            65,
            66,
            67,
            68,
            69,
            70,
            71,
            72,
            73,
            74,
            75,
            76,
            77,
            78,
            79,
            80,
            81,
            82,
            83,
            84,
            85,
            86,
            87,
            88,
            89,
            90,
            92,
            95,
            97,
            98,
            99,
            100,
            101,
            102,
            103,
            104,
            105,
            106,
            107,
            108,
            109,
            110,
            111,
            112,
            113,
            114,
            115,
            116,
            117,
            118,
            119,
            120,
            121,
            122,
            161,
            163,
            164,
            165,
            191,
            196,
            197,
            198,
            199,
            201,
            209,
            214,
            216,
            220,
            223,
            224,
            228,
            229,
            230,
            232,
            233,
            236,
            241,
            242,
            246,
            248,
            249,
            252,
            915,
            916,
            920,
            923,
            926,
            928,
            931,
            934,
            936,
            937,
        ]
        return gsm_7bit_map

    @classmethod
    def _get_added_gsm_7bit_ex_map(cls):
        added_gsm_7bit_ex_map = [12, 91, 92, 93, 94, 123, 124, 125, 126, 8364]
        return added_gsm_7bit_ex_map

    @classmethod
    def _text_to_unicode_pointcode_list(cls, plaintext):
        textlist = []
        for stg in plaintext:
            textlist.append(ord(stg))
        return textlist

    @classmethod
    def _detect_encoding(cls, plaintext):
        rf = cls._text_to_unicode_pointcode_list(plaintext)

        non_gsm_7bit_chars = set(rf) - set(cls._get_gsm_7bit_map())
        if not non_gsm_7bit_chars:
            return cls.GSM_7BIT

        non_gsm_7bit_ex_chars = non_gsm_7bit_chars - set(cls._get_added_gsm_7bit_ex_map())
        if not non_gsm_7bit_ex_chars:
            return cls.GSM_7BIT_EX

        return cls.UTF16

    @classmethod
    def count(cls, plaintext):
        textlist = cls._text_to_unicode_pointcode_list(plaintext)

        encoding = cls._detect_encoding(plaintext)
        length = len(textlist)

        if encoding == cls.GSM_7BIT_EX:
            exchars = [c for c in textlist if c in cls._get_added_gsm_7bit_ex_map()]
            lengthexchars = len(exchars)
            length += lengthexchars

        if encoding == cls.GSM_7BIT:
            permessage = cls.GSM_7BIT_LEN
            if length > cls.GSM_7BIT_LEN:
                permessage = cls.GSM_7BIT_LEN_MULTIPART
        elif encoding == cls.GSM_7BIT_EX:
            permessage = cls.GSM_7BIT_EX_LEN
            if length > cls.GSM_7BIT_EX_LEN:
                permessage = cls.GSM_7BIT_EX_LEN_MULTIPART
        else:
            permessage = cls.UTF16_LEN
            if length > cls.UTF16_LEN:
                permessage = cls.UTF16_LEN_MULTIPART

        # Convert the dividend to fload so the division will be a float number
        # and then convert the ceil result to int
        # since python 2.7 return a float
        messages = int(ceil(length / float(permessage)))
        remaining = (permessage * messages) - length

        returnset = {
            "encoding": encoding,
            "length": length,
            "per_message": permessage,
            "remaining": remaining,
            "messages": messages,
        }

        return returnset

    @classmethod
    def truncate(cls, plaintext, limitsms):
        count = cls.count(plaintext)

        if count.messages <= limitsms:
            return plaintext

        if count.encoding == "UTF16":
            limit = cls.UTF16_LEN

            if limitsms > 2:
                limit = cls.UTF16_LEN_MULTIPART

        if count.encoding != "UTF16":
            limit = cls.GSM_7BIT_LEN

            if limitsms > 2:
                limit = cls.GSM_7BIT_LEN_MULTIPART

        while True:
            text = plaintext[0 : limit * limitsms]  # noqa:  E203
            count = cls.count(plaintext)

            limit = limit - 1

            if count.messages < limitsms:
                break

        return text


class T2(ProviderIntegration):
    sms_from_number: dict
    username: str
    password: str

    endpoint_one = "send_sms"
    endpoint_multi = "send_multiple_sms"
    settings: object

    url = ""

    def __init__(self) -> None:
        super().__init__(settings=object())

    def ensure_credentials(self, settings: Optional[Settings] = None):
        if settings and getattr(settings, "TESTING", False):
            return
        self.username = getattr(settings, "NOTIFICATIONS_T2_USERNAME", None)
        self.password = getattr(settings, "NOTIFICATIONS_T2_PASSWORD", None)
        self.url = getattr(settings, "NOTIFICATIONS_SMS_API_URL", None)
        self.settings = settings
        assert self.username, "NOTIFICATIONS_T2_USERNAME is required"
        assert self.password, "NOTIFICATIONS_T2_PASSWORD is required"
        assert len(self.url) > 0, "NOTIFICATIONS_T2_PASSWORD is required"

    def client_send(self, sender: str, recipient: Recipient, msg: str, dlr_id: str):
        if not recipient.phone_number:
            return
        basic_auth = HTTPBasicAuth(self.username, self.password)
        response = requests.post(
            f"{self.url}{self.endpoint_one}",
            auth=basic_auth,
            json={"from_number": sender, "to_number": recipient.phone_number, "message": msg, "guid": dlr_id},
            verify=False,
            headers={"Content-Type": "application/json"},
            timeout=int(0.8 * NOTIFICATION_QUEABLE_HARD_TIME_LIMIT),
        )

        self.validate_send(response)

    def validate_send(self, response: object):
        assert response
        is_success(response.status_code)
        response_data = response.json()
        assert str(response_data["error_code"]) == "0"

    def parse_delivery_report(self, dlr: DeliveryReport):
        payload = json.loads(getattr(dlr, "payload", "{}"))
        dlr.status = (
            DeliveryReport.Status.DELIVERED
            if str(payload.get("status", "-1")) == str(DeliveryReport.Status.DELIVERED.value)
            else DeliveryReport.Status.NOT_DELIVERED
        )
        dlr.save(update_fields=["status"])

    @property
    def delivery_report_username_setting_name(self) -> str:
        return "t2-sms-dlr-user"

    @property
    def delivery_report_password_setting_name(self) -> str:
        return "t2-sms-dlr-password"

    def ensure_dlr_user(self, project_slug: str):
        if project_slug and (
            project := swapper.load_model("django_project_base", "Project").objects.filter(slug=project_slug).first()
        ):
            project_settings_model = swapper.load_model("django_project_base", "ProjectSettings")

            username_setting = project_settings_model.objects.filter(
                name=self.delivery_report_username_setting_name, project=project
            ).first()

            password_setting = project_settings_model.objects.filter(
                name=self.delivery_report_password_setting_name, project=project
            ).first()

            assert username_setting
            assert password_setting

            user, user_created = get_user_model().objects.get_or_create(
                username=username_setting.python_value,
                email="klemen.spruk@velis.si",
                first_name=username_setting.python_value,
                last_name=username_setting.python_value,
            )
            if user_created:
                user.set_password(password_setting.python_value)
                user.save()

            ProjectMember = swapper.load_model("django_project_base", "ProjectMember")
            ProjectMember.objects.get_or_create(member=user.userprofile, project=project)

    def enqueue_dlr_request(self, pk: str):
        pass

    def get_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return self._get_sms_message(notification)

    def _get_sms_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return super()._get_sms_message(notification)



import swapper

from django.db import connections
from django.test import TestCase
from dynamicforms import serializers
from rest_framework.test import APIRequestFactory

from django_project_base.base.fields import UserRelatedField
from django_project_base.base.filter_to_model import filter_queryset_or_model_to_project

# Note that the tests get created with demo app's models, not the base module's.
Project = swapper.load_model("django_project_base", "Project")
UserProfile = swapper.load_model("django_project_base", "Profile")
ProjectMember = swapper.load_model("django_project_base", "ProjectMember")


class FilterToModelTest(TestCase):
    def setUp(self):
        super().setUp()

        user = UserProfile.objects.create_user("test.user")
        project1 = Project.objects.create(name="project1", slug="project", owner=user)
        for i in range(10):
            u = UserProfile.objects.create_user(f"test.user.{i}")
            ProjectMember.objects.create(project=project1, member=u)

        user = UserProfile.objects.create_user("test1.user")
        project2 = Project.objects.create(name="project2", slug="project2", owner=user)
        for i in range(15):
            u = UserProfile.objects.create_user(f"test1.user.{i}")
            ProjectMember.objects.create(project=project2, member=u)

    def tearDown(self):
        Project.objects.all().delete()
        UserProfile.objects.all().delete()
        super().tearDown()

    def get_query_count(self):
        res = 0
        for c in connections:
            con = connections[c]
            con.force_debug_cursor = True
            res += len(con.queries)
        return res

    def test_project_selection(self):
        # this is basically a test of django-model-utils.InheritanceManager
        count = self.get_query_count()

        q = filter_queryset_or_model_to_project(queryset=None, model=ProjectMember,
                                                project=Project.objects.get(name="project1"))
        self.assertEqual(q.count(), 10)
        q = filter_queryset_or_model_to_project(queryset=ProjectMember.objects, model=None,
                                                project=Project.objects.get(name="project2"))
        self.assertEqual(q.count(), 15)

        # finally test if number of queries performed was as it needs to be
        self.assertEqual(self.get_query_count() - count, 4)

        q = filter_queryset_or_model_to_project(queryset=ProjectMember.objects, model=None, project=None)
        # Should return qs.none() because the selected project can't be determined from non-existing request here
        self.assertEqual(q.count(), 0)

        # finally test if number of queries performed was as it needs to be
        self.assertEqual(self.get_query_count() - count, 4)

    def context(self, project: "Project"):
        request = APIRequestFactory().get("/")
        if project:
            request.selected_project = project
        return {"request": request}

    def serializer(self, context):
        class TestSerializer(serializers.Serializer):
            user = UserRelatedField()

        return TestSerializer(context=context)

    def test_user_related_field_selection(self):
        ser = self.serializer(context=self.context(Project.objects.get(name="project1")))
        field = ser.fields["user"]
        self.assertIsNotNone(field)
        self.assertEqual(field.get_queryset().count(), 10)

        ser = self.serializer(context=self.context(None))
        field = ser.fields["user"]
        self.assertIsNotNone(field)
        self.assertGreaterEqual(field.get_queryset().count(), 25)



import copy
import datetime

from abc import ABC, abstractmethod
from gettext import gettext

import swapper

from django.conf import settings
from django.shortcuts import get_object_or_404
from rest_registration.settings import registration_settings

from django_project_base.constants import EMAIL_SENDER_ID_SETTING_NAME, SMS_SENDER_ID_SETTING_NAME
from django_project_base.notifications.email_notification import SystemEMailNotificationWithListOfEmails
from django_project_base.notifications.models import DjangoProjectBaseMessage


class UserModel:
    pk: int


class BaseEvent(ABC):
    user: UserModel

    def __init__(self, user: UserModel) -> None:
        super().__init__()
        self.user = user

    @abstractmethod
    def trigger(self, payload=None, **kwargs):
        pass

    @abstractmethod
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        self.trigger(payload=payload, **kwargs)


class ProjectSettingChangedEvent(BaseEvent):
    def trigger(self, payload=None, **kwargs):
        super().trigger(payload=payload, **kwargs)

    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)


class EmailSenderChangedEvent(ProjectSettingChangedEvent):
    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)

    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state, new_state, payload, **kwargs)

        if new_state.name == EMAIL_SENDER_ID_SETTING_NAME:
            # TODO: THIS IS ONLY FOR AWS FOR NOW
            from django_project_base.aws.ses import AwsSes

            if not old_state:
                AwsSes.add_sender_email(new_state.python_value)
            elif (old_state.python_value != new_state.python_value) or (
                new_state.python_value
                and new_state.pending_value
                and new_state.python_pending_value != new_state.python_value
            ):
                AwsSes.add_sender_email(new_state.python_pending_value)

            project_settings_manager = swapper.load_model("django_project_base", "ProjectSettings").objects
            for sender in set(AwsSes.list_sender_emails()) - (
                set(project_settings_manager.filter(name=EMAIL_SENDER_ID_SETTING_NAME).values_list("value", flat=True))
                | set(
                    project_settings_manager.filter(name=EMAIL_SENDER_ID_SETTING_NAME).values_list(
                        "pending_value", flat=True
                    )
                )
            ):
                AwsSes.remove_sender_email(sender)


class SmsSenderChangedEvent(ProjectSettingChangedEvent):
    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)

    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state, new_state, payload, **kwargs)

        if new_state.name == SMS_SENDER_ID_SETTING_NAME:
            pass


class UserInviteFoundEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload=payload)
        if not payload or payload.accepted or not kwargs.get("request"):
            return

        swapper.load_model("django_project_base", "ProjectMember").objects.get_or_create(
            project=payload.project, member=self.user.userprofile
        )
        from django_project_base.account.rest.project_profiles import ProjectProfilesViewSet

        setattr(kwargs["request"], "selected_project", payload.project)
        ProjectProfilesViewSet().save_club_member_data(kwargs["request"], self.user.userprofile)
        payload.accepted = datetime.datetime.now()
        payload.save(update_fields=["accepted"])

        if current_project_attr := (
            getattr(settings, "DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES", {})
            .get("project", {})
            .get("value_name", None)
        ):
            kwargs["request"].session[current_project_attr] = payload.project.slug
        kwargs["request"].session.pop("invite-pk", None)


class UserRegisteredEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)
        if not payload:
            return

        if invite_pk := payload.session.get("invite-pk"):
            invite = get_object_or_404(swapper.load_model("django_project_base", "Invite"), pk=invite_pk)
            UserInviteFoundEvent(self.user).trigger(payload=invite, request=payload)
            return
        payload.session.pop("invite-pk", None)
        if registration_settings.REGISTER_VERIFICATION_ENABLED:
            registration_settings.REGISTER_VERIFICATION_EMAIL_SENDER(request=payload, user=self.user)


class UserLoginEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)
        if not payload:
            return

        if invite_pk := payload.session.get("invite-pk"):
            invite = get_object_or_404(swapper.load_model("django_project_base", "Invite"), pk=invite_pk)
            UserInviteFoundEvent(self.user).trigger(payload=invite, request=payload)
            return
        payload.session.pop("invite-pk", None)


class ProjectSettingConfirmedEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)
        if not payload:
            return
        from django_project_base.aws.ses import AwsSes

        def confirm(item):
            item.value = copy.copy(item.python_pending_value)
            item.pending_value = None
            item.save(update_fields=["value", "pending_value"])

        # not self.user Event is trigerred from management command
        if payload.name == EMAIL_SENDER_ID_SETTING_NAME and (
            payload.python_pending_value in AwsSes.list_verified_sender_emails() or not self.user
        ):
            confirm(payload)
        if payload.name == SMS_SENDER_ID_SETTING_NAME:
            if not self.user:
                confirm(payload)


class ProjectSettingActionRequiredEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)
        if not payload:
            return

        if to := getattr(settings, "ADMINS", getattr(settings, "MANAGERS", [])):
            SystemEMailNotificationWithListOfEmails(
                message=DjangoProjectBaseMessage(
                    subject=gettext("Project settings action required"),
                    body=f"{gettext('Action required for setting')} {payload.name} in project {payload.project.name}",
                    footer="",
                    content_type=DjangoProjectBaseMessage.HTML,
                ),
                recipients=to,
            ).send()


class ProjectSettingPendingResetEvent(BaseEvent):
    def trigger_changed(self, old_state=None, new_state=None, payload=None, **kwargs):
        super().trigger_changed(old_state=old_state, new_state=new_state, payload=payload, **kwargs)

    def trigger(self, payload=None, **kwargs):
        super().trigger(payload, **kwargs)
        if not payload:
            return
        from django_project_base.aws.ses import AwsSes

        pending_value = copy.copy(payload.python_pending_value)
        payload.pending_value = None
        payload.save(update_fields=["pending_value"])

        if payload.name == EMAIL_SENDER_ID_SETTING_NAME:
            if pending_value in AwsSes.list_sender_emails():
                AwsSes.remove_sender_email(pending_value)
            AwsSes.add_sender_email(payload.python_value)
            if payload.python_value not in AwsSes.list_verified_sender_emails():
                payload.action_required = True
                payload.save(update_fields=["action_required"])
        if payload.name == SMS_SENDER_ID_SETTING_NAME:
            payload.action_required = True
            payload.save(update_fields=["action_required"])



# Generated by Django 4.2.4 on 2023-08-30 11:18

from django.db import migrations, models
import django_project_base.licensing.models


class Migration(migrations.Migration):
    replaces = [
        ("licensing", "0002_remove_licenseaccessuse_comment_and_more"),
        ("licensing", "0003_alter_licenseaccessuse_comment_and_more"),
        ("licensing", "0004_alter_licenseaccessuse_amount"),
    ]

    dependencies = [
        ("licensing", "0001_initial"),
    ]

    operations = [
        migrations.AlterField(
            model_name="licenseaccessuse",
            name="type",
            field=models.IntegerField(
                choices=[
                    (
                        django_project_base.licensing.models.LicenseAccessUse.UseType["USE"],
                        "Used access",
                    )
                ],
                default=django_project_base.licensing.models.LicenseAccessUse.UseType["USE"],
                verbose_name="Type",
            ),
        ),
        migrations.AlterField(
            model_name="licenseaccessuse",
            name="comment",
            field=models.JSONField(null=True, verbose_name="Comment"),
        ),
        migrations.AlterField(
            model_name="licenseaccessuse",
            name="type",
            field=models.IntegerField(
                choices=[
                    (
                        django_project_base.licensing.models.LicenseAccessUse.UseType["USE"],
                        "Used access",
                    ),
                    (
                        django_project_base.licensing.models.LicenseAccessUse.UseType["ADMIN_USE"],
                        "Used access by admin",
                    ),
                ],
                default=django_project_base.licensing.models.LicenseAccessUse.UseType["USE"],
                verbose_name="Type",
            ),
        ),
        migrations.AddField(
            model_name="licenseaccessuse",
            name="amount",
            field=models.FloatField(default=0),
        ),
    ]



from typing import Any, Dict, Optional, Union

from django.contrib.auth import get_user_model
from django.db.models import Model, Q
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import (
    extend_schema,
    extend_schema_field,
    extend_schema_view,
    OpenApiParameter,
    OpenApiResponse,
    OpenApiTypes,
)
from dynamicforms import fields, serializers, viewsets
from dynamicforms.action import Actions, FormButtonAction, FormButtonTypes
from hijack.helpers import login_user, release_hijack
from rest_framework import status
from rest_framework.exceptions import PermissionDenied
from rest_framework.permissions import IsAdminUser, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response

import django_project_base.base.fields


class ImpersonateRequestSerializer(serializers.Serializer):
    """
    Impersonation options. Fill out any of the fields to identify the impersonated user
    """

    id = fields.IntegerField(required=False, label=_("User ID"))
    email = fields.EmailField(required=False, label=_("User email"))
    username = fields.CharField(required=False)


@extend_schema_field(OpenApiTypes.NONE)
class ImpersonateUserIdField(django_project_base.base.fields.UserRelatedField):
    def __init__(
        self,
        queryset_filter: Optional[Union[Q, Dict[str, Any]]] = None,
        queryset_exclude: Optional[Union[Q, Dict[str, Any]]] = None,
        **kwargs,
    ):
        kwargs.setdefault("placeholder", _("Select a user to impersonate"))
        kwargs.setdefault("label", _("User"))
        kwargs.setdefault("required", False)
        kwargs.setdefault("allow_null", True)
        super().__init__(queryset_filter, queryset_exclude, **kwargs)
        self.filter_selected_project = False


class ImpersonateUserDialogSerializer(serializers.Serializer):
    template_context = dict(
        url_reverse="profile-base-impersonate-user", url_reverse_kwargs=None, dialog_header_classes="bg-info"
    )
    form_titles = {"new": _("Impersonate a user")}

    id = ImpersonateUserIdField()

    actions = Actions(
        FormButtonAction(btn_type=FormButtonTypes.CANCEL, name="cancel", label=_("Cancel")),
        FormButtonAction(btn_type=FormButtonTypes.SUBMIT, name="submit", label=_("Impersonate")),
        add_form_buttons=False,
    )


@extend_schema_view(
    retrieve=extend_schema(
        description="Retrieves dialog definition for impersonation dialog",
        parameters=[OpenApiParameter("id", OpenApiTypes.STR, OpenApiParameter.PATH, enum=["new"])],
    ),
)
class ImpersonateUserViewset(viewsets.SingleRecordViewSet):
    serializer_class = ImpersonateUserDialogSerializer

    def get_permissions(self):
        if self.action == "destroy":
            return [
                IsAuthenticated(),
            ]
        elif self.action == "update":
            return [
                IsAdminUser(),
            ]
        elif self.action == "create":
            return [
                IsAdminUser(),
            ]
        return super().get_permissions()

    def new_object(self):
        return dict(id=None)

    # noinspection PyMethodMayBeStatic
    def __validate(self, req_data: dict) -> dict:
        ser: ImpersonateRequestSerializer = ImpersonateRequestSerializer(data=req_data)
        ser.is_valid(raise_exception=True)
        return ser.validated_data

    class ImpersonateUser(serializers.Serializer):
        id = fields.IntegerField(required=True, help_text=_("Target user pk"))

    @extend_schema(
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Forbidden"),
        },
        description="Logout as another user",
    )
    def destroy(self, request: Request) -> Response:
        release_hijack(request)
        return Response()

    @extend_schema(exclude=True)
    def update(self, request: Request, *args, **kwargs) -> Response:
        self._hijack(request=request)
        return Response()

    def _hijack(self, request: Request):
        validated_data: dict = self.__validate(request.data)
        hijacked_user: Model = get_object_or_404(get_user_model(), **validated_data)
        if request.user == hijacked_user:
            raise PermissionDenied(_("Impersonating self is not allowed"))
        login_user(request, hijacked_user)

    @extend_schema(
        request=ImpersonateUser,
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(
                description="Forbidden. You do not have permission to perform this action or "
                "Impersonating self is not allowed"
            ),
            status.HTTP_404_NOT_FOUND: OpenApiResponse(description="User matching provided data not found"),
        },
        description="Login as another user and work on behalf of other user without having to know their credentials",
    )
    def create(self, request: Request, *args, **kwargs) -> Response:
        self._hijack(request=request)
        return Response()



from typing import Optional

from django_redis import get_redis_connection

from django_project_base.caching.cache_queue import CacheQueue


class CacheQueueRedis(CacheQueue):
    def set_cache(self):
        self.cache = get_redis_connection(self.cache_name)

    def rpush(self, *values):
        self.cache.rpush(self.key, *values)
        self.update_timeout()

    def lpush(self, *values):
        self.cache.lpush(self.key, *values)
        self.update_timeout()

    def rpop(self, count: Optional[int] = None):
        ret = self.cache.rpop(self.key, count)
        self.update_timeout()
        return ret

    def lpop(self, count: Optional[int] = None):
        ret = self.cache.lpop(self.key, count)
        self.update_timeout()
        return ret

    def lrange(self, count=-1):
        return self.cache.lrange(self.key, 0, count)

    def ltrim(self, count=0):
        self.cache.ltrim(self.key, count, -1)
        self.update_timeout()

    def update_timeout(self):
        if self.timeout is None:
            self.cache.persist(self.key)
        else:
            self.cache.expire(self.key, self.timeout)



import socket

import swapper
from django.urls import reverse
from rest_framework import status
from rest_framework.response import Response

from django_project_base.notifications.tests.notifications_transaction_test_case import (
    NotificationsTransactionTestCase,
    TestNotificationViaEmail,
)
from django_project_base.utils import get_pk_name


class TestRetrieveMail(NotificationsTransactionTestCase):
    def test_retrieve_mail(self):
        mail: TestNotificationViaEmail = self.create_notification_email_object()
        notification = mail.send()
        response: Response = self.api_client.get(
            reverse("notification-detail", kwargs=dict(pk=notification.pk)),
            format="json",
            HTTP_HOST=socket.gethostname().lower(),
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self._login_to_api_client_with_test_user()
        response: Response = self.api_client.get(
            reverse("notification-detail", kwargs=dict(pk=notification.pk)),
            format="json",
            HTTP_HOST=socket.gethostname().lower(),
            HTTP_CURRENT_PROJECT=swapper.load_model("django_project_base", "Project").objects.first().slug,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.assertIsNotNone(response.data)
        self.assertIsNotNone(response.data[get_pk_name(notification)], str(notification.pk))
        notification.refresh_from_db()
        self.assertEqual(1, notification.counter)



from django_project_base.account.social_auth.settings import SOCIAL_AUTH_SETTINGS

ACCOUNT_SETTINGS = []

for set in SOCIAL_AUTH_SETTINGS:
    ACCOUNT_SETTINGS.append(set)

ACCOUNT_SETTINGS = tuple(ACCOUNT_SETTINGS)



import socket

import swapper
from django.urls import reverse
from rest_framework import status
from rest_framework.response import Response

from django_project_base.notifications.base.channels.mail_channel import MailChannel
from django_project_base.notifications.email_notification import EMailNotification
from django_project_base.notifications.models import DjangoProjectBaseMessage, DjangoProjectBaseNotification
from django_project_base.notifications.tests.notifications_transaction_test_case import NotificationsTransactionTestCase


class TestCreateMail(NotificationsTransactionTestCase):
    def test_send_mail(self):
        self.assertIsNotNone(
            EMailNotification(
                message=DjangoProjectBaseMessage(subject="test", body="", footer=""),
                raw_recipents=[
                    self.test_user.pk,
                ],
                project=swapper.load_model("django_project_base", "Project").objects.first().slug,
                recipients=[
                    self.test_user.pk,
                ],
            ).send()
        )
        self.assertEqual(DjangoProjectBaseNotification.objects.all().count(), 1)
        self.assertIsNotNone(
            EMailNotification(
                message=DjangoProjectBaseMessage(subject="test", body="", footer=""),
                raw_recipents=[
                    self.test_user.pk,
                ],
                project=swapper.load_model("django_project_base", "Project").objects.first().slug,
                recipients=[
                    self.test_user.pk,
                ],
                persist=False,
            ).send()
        )
        self.assertEqual(DjangoProjectBaseNotification.objects.all().count(), 1)

    def test_api_create_mail(self):
        self._login_to_api_client_with_test_user()
        response: Response = self.api_client.put(
            reverse("notification-detail", kwargs=dict(pk="new")),
            format="json",
            data={
                "message_body": "demo",
                "message_subject": "demo",
                "message_to": [self.test_user.pk],
                "send_on_channels": [MailChannel.name],
            },
            HTTP_HOST=socket.gethostname().lower(),
            HTTP_CURRENT_PROJECT=swapper.load_model("django_project_base", "Project").objects.first().slug,
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)



import json

import swapper

from django.core.cache import cache
from django.shortcuts import get_object_or_404
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view, OpenApiParameter, OpenApiResponse, OpenApiTypes
from dynamicforms import fields as df_fields, serializers as df_serializers, viewsets as df_viewsets
from dynamicforms.action import Actions
from rest_framework import fields, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.exceptions import ValidationError
from rest_framework.permissions import AllowAny, IsAdminUser, IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_registration.api.views import reset_password, send_reset_password_link, verify_email
from rest_registration.exceptions import UserNotFound
from rest_registration.settings import registration_settings

from django_project_base.account.constants import RESET_USER_PASSWORD_VERIFICATION_CODE


class ResetPasswordSerializer(serializers.Serializer):
    user_id = fields.CharField(required=True)
    timestamp = fields.CharField(required=True)
    signature = fields.CharField(required=True)
    password = fields.CharField(required=True)


class ResetPasswordCodeSerializer(serializers.Serializer):
    user_id = fields.CharField(required=True)
    code = fields.CharField(required=True)

    def validate_code(self, value) -> str:
        if value and value == cache.get(RESET_USER_PASSWORD_VERIFICATION_CODE + str(self.initial_data.get("user_id"))):
            return value
        raise ValidationError("Invalid code")


class ResetPasswordViewSet(viewsets.ViewSet):
    serializer_class = ResetPasswordSerializer
    permission_classes = (AllowAny,)

    @extend_schema(
        description="Reset password through a link sent in email, given the signature and timestamp from the link.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK Reset link sent"),
            status.HTTP_400_BAD_REQUEST: OpenApiResponse(description="Bad request."),
            status.HTTP_404_NOT_FOUND: OpenApiResponse(description="Reset password verification disabled"),
        },
    )
    @action(detail=False, methods=["post"], url_path="reset-password", url_name="reset-password")
    def reset_password(self, request: Request) -> Response:
        code_ser = ResetPasswordCodeSerializer(
            data=json.loads(request._request.body.decode()), context=dict(request=request)
        )
        code_ser.is_valid(raise_exception=True)
        response = reset_password(request._request)
        cache.delete(RESET_USER_PASSWORD_VERIFICATION_CODE + str(code_ser["user_id"]))
        return response

    @extend_schema(
        description="Verify email via signature.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK Email verified successfully"),
            status.HTTP_400_BAD_REQUEST: OpenApiResponse(description="Bad request"),
        },
    )
    @action(detail=False, methods=["post"], url_path="verify-email", url_name="verify-email")
    def verify_email(self, request: Request) -> Response:
        return verify_email(request._request)


class ResetPasswordAdminSerializer(df_serializers.Serializer):
    template_context = dict(
        url_reverse="admin-invalidate-password", dialog_classes="modal-lg", dialog_header_classes="bg-info"
    )
    form_titles = {
        "table": "",
        "new": _("Invalidate password"),
        "edit": "",
    }

    user_id = df_fields.IntegerField(required=True, help_text=_("Target user id"))

    actions = Actions()


@extend_schema_view(
    retrieve=extend_schema(parameters=[OpenApiParameter("id", OpenApiTypes.STR, OpenApiParameter.PATH, enum=["new"])]),
)
class InvalidatePasswordAdminViewSet(df_viewsets.SingleRecordViewSet):
    serializer_class = ResetPasswordAdminSerializer

    permission_classes = (IsAuthenticated, IsAdminUser)  # TODO: permission should be based on project role

    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)
        request._dont_enforce_csrf_checks = True
        return request

    def new_object(self):
        return dict(user_id="")

    @extend_schema(
        description="Marks the user password to be changed."
        "When user logs to an app UI user is shown change password dialog.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    def create(self, request: Request) -> Response:
        self.serializer_class(data=request.data).is_valid(raise_exception=True)
        profile_obj = get_object_or_404(
            swapper.load_model("django_project_base", "Profile"), user_ptr_id=request.data.get("user_id")
        )
        profile_obj.password_invalid = True
        profile_obj.save(update_fields=["password_invalid"])
        return Response()


class SendResetPasswordLinkSerializer(serializers.Serializer):
    login = fields.CharField(required=True)


class SendResetPasswordLinkViewSet(viewsets.ViewSet):
    serializer_class = SendResetPasswordLinkSerializer()
    permission_classes = (AllowAny,)

    @extend_schema(
        description="Send email with reset password link.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(
                description="OK Reset link sent", response=SendResetPasswordLinkSerializer()
            ),
            status.HTTP_400_BAD_REQUEST: OpenApiResponse(description="Bad request"),
            status.HTTP_404_NOT_FOUND: "Reset password verification disabled",
        },
    )
    @action(detail=False, methods=["post"], url_path="send-reset-password-link", url_name="send-reset-password-link")
    def send_reset_password_link(self, request: Request) -> Response:
        try:
            req_data = request._request.body
            send_reset_password_link(request._request)
            reset_data = registration_settings.RESET_PASSWORD_VERIFICATION_EMAIL_SENDER(
                request=request,
                user=registration_settings.SEND_RESET_PASSWORD_LINK_USER_FINDER(json.loads(req_data.decode())),
                send=True,
            )
            return Response(reset_data)
        except UserNotFound:
            return Response()
        except Exception as e:
            raise e



from typing import List

import boto3

from django.conf import settings
from rest_framework import status


def get_aws_session():
    if settings.TESTING:

        class FakeSession:
            client = lambda t, i: object()  # noqa: E731

        return FakeSession()
    return boto3.Session(
        aws_access_key_id=getattr(settings, "NOTIFICATIONS_AWS_SES_ACCESS_KEY_ID", None),
        aws_secret_access_key=getattr(settings, "NOTIFICATIONS_AWS_SES_SECRET_ACCESS_KEY", None),
        region_name=getattr(settings, "NOTIFICATIONS_AWS_SES_REGION_NAME", None),
    )


class AwsSes:
    client = (lambda: get_aws_session().client("ses"))()

    @staticmethod
    def list_sender_emails() -> List[str]:
        return AwsSes.client.list_identities()["Identities"]

    @staticmethod
    def list_verified_sender_emails() -> List[str]:
        return AwsSes.client.list_verified_email_addresses()["VerifiedEmailAddresses"]

    @staticmethod
    def remove_sender_email(email: str):
        if email in AwsSes.list_sender_emails():
            # TODO: EACH DEPLOYMENT SHOULD HAVE ITS OWN AWS ACCOUNT FOR THIS TO BE ENABLED
            # TODO: ALSO DEV MACHINES SHOULD HAVE ITS OWN ACCOUNT
            pass
            # assert (
            #     AwsSes.client.delete_identity(
            #         Identity=email,
            #     )[
            #         "ResponseMetadata"
            #     ]["HTTPStatusCode"]
            #     == status.HTTP_200_OK
            # )

    @staticmethod
    def add_sender_email(email: str):
        if email not in AwsSes.list_sender_emails():
            assert (
                AwsSes.client.verify_email_identity(EmailAddress=email)["ResponseMetadata"]["HTTPStatusCode"]
                == status.HTTP_200_OK
            )



import logging

from django.conf import Settings
from django.utils import timezone

from django_project_base.constants import NOTIFICATION_QUEUE_NAME
from django_project_base.notifications.base.enums import ChannelIdentifier
from django_project_base.notifications.models import DjangoProjectBaseNotification


class SendNotificationService(object):
    settings: Settings
    use_default_db_connection = False

    def __init__(self, settings: Settings, use_default_db_connection=False) -> None:
        super().__init__()
        self.settings = settings
        self.use_default_db_connection = use_default_db_connection

    def make_send(
        self, notification: DjangoProjectBaseNotification, extra_data, resend=False
    ) -> DjangoProjectBaseNotification:
        # TODO: THIS SHOULD BE CALLED ONLY FROM CELERY BACKGROUND TASK

        sent_channels: list = []
        failed_channels: list = []
        db_name = NOTIFICATION_QUEUE_NAME if not self.use_default_db_connection else "default"

        exceptions = ""
        from django_project_base.licensing.logic import LogAccessService

        if notification.required_channels is None:
            return notification
        already_sent_channels = set(
            filter(
                lambda i: i not in (None, "") and i,
                notification.sent_channels.split(",") if notification.sent_channels else [],
            )
        )
        required_channels = set(
            filter(
                lambda i: i not in (None, "") and i,
                notification.required_channels.split(",") if notification.required_channels else [],
            )
        ) - (already_sent_channels if not resend else set())
        already_failed_channels = set(
            filter(
                lambda i: i not in (None, "") and i,
                notification.failed_channels.split(",") if notification.failed_channels else [],
            )
        )

        from django_project_base.notifications.base.channels.sms_channel import SmsChannel

        if notification.send_notification_sms:
            required_channels.add(SmsChannel.name)

        for channel_identifier in required_channels:
            channel = ChannelIdentifier.channel(
                channel_identifier,
                settings=self.settings,
                project_slug=notification.project_slug,
                ensure_dlr_user=False,
            )
            try:
                # check license
                any_sent = LogAccessService(db=db_name).log(
                    user_profile_pk=notification.user,
                    notifications_channels_state=sent_channels,
                    record=notification,
                    item_price=channel.notification_price,
                    comment=str(channel),
                    on_sucess=lambda: channel.send(
                        notification=notification, extra_data=extra_data, settings=self.settings
                    ),
                    is_system_notification=extra_data.get("is_system_notification"),
                    sender=channel.sender(notification),
                )
                sent_channels.append(channel) if any_sent > 0 else failed_channels.append(channel)
            except Exception as e:
                logging.getLogger(__name__).error(e)
                failed_channels.append(channel)
                exceptions += f"{str(e)}\n\n"

        if notification.created_at:
            if required_channels:
                notification.sent_channels = (
                    ",".join(
                        set(
                            list(
                                map(
                                    lambda f: str(f),
                                    filter(
                                        lambda d: d is not None,
                                        map(lambda c: c.name, sent_channels),
                                    ),
                                )
                            )
                            + list(already_sent_channels)
                        )
                    )
                    if sent_channels
                    else ",".join(already_sent_channels)
                )
            notification.failed_channels = (
                ",".join(
                    set(
                        list(
                            map(
                                lambda f: str(f),
                                filter(
                                    lambda d: d is not None,
                                    map(lambda c: c.name, failed_channels),
                                ),
                            )
                        )
                        + list(already_failed_channels)
                    )
                )
                if failed_channels
                else ",".join(already_failed_channels)
            )
            notification.sent_at = timezone.now().timestamp() if notification.sent_channels else None
            notification.exceptions = exceptions if exceptions else None

            _req = notification.required_channels.split(",") if notification.required_channels else []
            _snt = notification.sent_channels.split(",") if notification.sent_channels else []
            if set(_snt) == set(_req):
                notification.failed_channels = ""

            notification.save(
                update_fields=[
                    "sent_at",
                    "sent_channels",
                    "failed_channels",
                    "exceptions",
                ],
                using=db_name,
            )
        return notification



import datetime

from django.conf import settings

from django_project_base.celery.background_tasks.notification_tasks import send_notification_task
from django_project_base.celery.settings import NOTIFICATION_SEND_PAUSE_SECONDS, NOTIFICATIONS_QUEUE_VISIBILITY_TIMEOUT
from django_project_base.constants import NOTIFICATION_QUEUE_NAME
from django_project_base.notifications.models import DjangoProjectBaseNotification


class QueableNotificationMixin(object):
    def enqueue_notification(self, notification: DjangoProjectBaseNotification, extra_data):
        if getattr(settings, "TESTING", False):
            return
        now_ts: int = int(datetime.datetime.now().timestamp())
        if notification.delayed_to - now_ts < NOTIFICATION_SEND_PAUSE_SECONDS:
            send_notification_task.apply_async(
                (notification, extra_data), queue=NOTIFICATION_QUEUE_NAME, serializer="pickle"
            )  # type: ignore
            return
        delay: int = notification.delayed_to - now_ts
        if delay > NOTIFICATIONS_QUEUE_VISIBILITY_TIMEOUT:
            raise Exception("Notification delay exceeds permitted delay")
        send_notification_task.apply_async(
            (notification, extra_data),
            eta=datetime.datetime.utcnow() + datetime.timedelta(seconds=delay),
            queue=NOTIFICATION_QUEUE_NAME,
            serializer="pickle",
        )  # type: ignore



from django.conf import settings

from django_project_base.licensing.logic import LogAccessService
from django_project_base.notifications.base.enums import ChannelIdentifier
from django_project_base.notifications.tests.notifications_transaction_test_case import NotificationsTransactionTestCase


class TestRemainingLicense(NotificationsTransactionTestCase):
    def test_remaining_license(self):
        mail = self.create_notification_email_object().send()
        log_service = LogAccessService()
        usage = log_service.report(user=self.test_user)
        credit = log_service._user_access_user_inflow(self.test_user.pk)
        channel = ChannelIdentifier.channel(
            mail.required_channels.split(",")[0], ensure_dlr_user=False, settings=settings
        )
        self.assertEqual(credit - channel.notification_price, usage["remaining_credit"])



from typing import List, Optional

from django.contrib.contenttypes.models import ContentType
from django.db import connections
from django.db.models import Model, Sum
from django.utils.translation import gettext
from dynamicforms import fields
from dynamicforms.serializers import Serializer
from rest_framework.exceptions import PermissionDenied

from django_project_base.licensing.models import LicenseAccessUse


class LicenseUsageReport(Serializer):
    item = fields.CharField(read_only=True)
    usage_sum = fields.FloatField(read_only=True)


class LicenseReportSerializer(Serializer):
    credit = fields.FloatField(read_only=True)
    used_credit = fields.FloatField(read_only=True)
    remaining_credit = fields.FloatField(read_only=True)

    usage_report = fields.ListField(child=LicenseUsageReport(), allow_empty=True, read_only=True)


class LogAccessService:
    db = "default"

    def __init__(self, db: Optional[str] = None) -> None:
        super().__init__()
        if db:
            self.db = db

    def _user_access_user_inflow(self, user_id) -> float:
        return abs(
            LicenseAccessUse.objects.using(self.db)
            .filter(user_id=str(user_id), amount__lt=0)
            .aggregate(Sum("amount"))
            .get("amount__sum", None)
            or 0
        )

    def report(self, user: Model) -> dict:
        user_query = LicenseAccessUse.objects.using(self.db).filter(
            user_id=str(user.pk), amount__isnull=False, amount__gt=0
        )

        usage_report = []
        added_types = []
        for agg in user_query.values("content_type").annotate(count=Sum("amount")):
            if (
                content_type := ContentType.objects.using(self.db).get(pk=agg["content_type"])
            ) and content_type.model_class()._meta.verbose_name not in added_types:
                usage_report.append(
                    {"item": content_type.model_class()._meta.verbose_name, "usage_sum": round(agg.get("count", 0), 2)}
                )
                added_types.append(content_type.model_class()._meta.verbose_name)
        used_credit = 0
        if user_query.exists():
            used_credit = round(user_query.aggregate(count=Sum("amount")).get("count", 0), 4)

        credit = self._user_access_user_inflow(user.pk)

        return LicenseReportSerializer(
            {
                "credit": round(credit, 2),
                "used_credit": used_credit,
                "usage_report": usage_report,
                "remaining_credit": round(credit - used_credit, 4),
            }
        ).data

    def log(
        self,
        user_profile_pk,
        notifications_channels_state: List[str],
        record: Model,
        item_price: float,
        comment: str,
        on_sucess=None,
        **kwargs,
    ) -> int:
        connections["default"] = connections[self.db]
        content_type = ContentType.objects.get_for_model(model=record._meta.model)
        used = (
            LicenseAccessUse.objects.using(self.db)
            .filter(user_id=str(user_profile_pk), content_type=content_type)
            .aggregate(Sum("amount"))
            .get("amount__sum", None)
            or 0
        )
        if notifications_channels_state:
            items = {i: notifications_channels_state.count(i) for i in notifications_channels_state}
            from django_project_base.notifications.base.enums import ChannelIdentifier

            chl_prices = {i.name: i.notification_price for i in ChannelIdentifier.supported_channels()}
            for used_channel in list(set(list(items.keys()))):
                used += chl_prices.get(used_channel, 0) * items.get(used_channel, 0)

        if not kwargs.get("is_system_notification") and used >= 0:
            raise PermissionDenied(gettext("Your license is consumed. Please contact support."))

        if on_sucess:
            accesses_used = on_sucess()
        else:
            accesses_used = 1

        amount = accesses_used * item_price

        LicenseAccessUse.objects.using(self.db).create(
            type=LicenseAccessUse.UseType.USE,
            user_id=str(user_profile_pk),
            content_type_object_id=str(record.pk).replace("-", ""),
            content_type=content_type,
            amount=amount,
            comment=dict(comment=comment, count=accesses_used, item_price=item_price, sender=kwargs.get("sender", "")),
        )

        return accesses_used



import base64
import re

from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from typing import List, Optional, Union

import boto3

from django.conf import Settings
from rest_framework.status import is_success

from django_project_base.notifications.base.channels.channel import Recipient
from django_project_base.notifications.base.channels.integrations.provider_integration import ProviderIntegration
from django_project_base.notifications.models import (
    DeliveryReport,
    DjangoProjectBaseMessage,
    DjangoProjectBaseNotification,
)


class AwsSes(ProviderIntegration):
    key_id: str
    access_key: str
    region: str

    is_sms_provider = False

    def __init__(self) -> None:
        super().__init__(settings=object())

    def validate_send(self, response: dict):
        assert response
        assert is_success(response.get("ResponseMetadata", {}).get("HTTPStatusCode", 500))

    def _get_sms_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return super()._get_sms_message(notification)

    def ensure_credentials(self, settings: Optional[Settings] = None):
        if settings and getattr(settings, "TESTING", False):
            return
        self.key_id = getattr(settings, "NOTIFICATIONS_AWS_SES_ACCESS_KEY_ID", None)
        self.access_key = getattr(settings, "NOTIFICATIONS_AWS_SES_SECRET_ACCESS_KEY", None)
        self.region = getattr(settings, "NOTIFICATIONS_AWS_SES_REGION_NAME", None)
        self.settings = settings
        assert self.key_id, "AWS SES key id required"
        assert self.access_key, "AWS SES key id access key required"
        assert self.region, "AWS SES region required"

    def get_message(self, notification: DjangoProjectBaseNotification) -> dict:
        msg = {
            "Body": {},
            "Subject": {
                "Charset": "UTF-8",
                "Data": str(notification.message.subject),
            },
        }
        msg["Body"][
            "Html" if notification.message.content_type.lower() == DjangoProjectBaseMessage.HTML else "Text"
        ] = {
            "Charset": "UTF-8",
            "Data": str(notification.message.body),
        }
        return msg

    def parse_delivery_report(self, dlr: DeliveryReport):
        pass

    @property
    def delivery_report_username_setting_name(self) -> str:
        return "aws-ses-email-dlr-user"

    @property
    def delivery_report_password_setting_name(self) -> str:
        return "aws-ses-email-dlr-password"

    def ensure_dlr_user(self, project_slug: str):
        pass

    def enqueue_dlr_request(self, pk: str):
        DeliveryReport.objects.filter(pk=pk).update(status=DeliveryReport.Status.DELIVERED)

    def parse_msg_images(self, msg: dict) -> MIMEMultipart:
        mail = MIMEMultipart("mixed")
        mail["Subject"] = msg["Subject"]["Data"]

        # extract embedded images
        pattern = r'<img[^>]*src="(data:image/([^;]+);base64,([^"]+))"[^>]*>'
        content: str = msg["Body"]["Html"]["Data"]
        # make attachements from images and replace
        embedded: List[str, MIMEImage] = []
        for index, (src, type, data) in enumerate(re.findall(pattern, content)):
            img_name = f"image{index + 1}"
            img_temp = MIMEImage(base64.b64decode(data), type)
            img_temp.add_header("Content-ID", f"{img_name}")
            # this looks kinda ugly, but we need all this data to properly deduplicate and write images in
            embedded.append((src, (img_name, img_temp)))

        deduplicated_embedded = dict(embedded).items()

        # deduplicate images, replace and register attachments
        for src, (img_name, img) in deduplicated_embedded:
            content = content.replace(src, f"cid:{img_name}")

        # attach body
        mail.attach(MIMEText(content, "html"))

        # just to make sure, body comes first, then image
        for _, (_, img) in deduplicated_embedded:
            mail.attach(img)

        return mail

    def client_send(self, sender: str, recipient: Recipient, msg: dict, dlr_id: str):
        if not recipient.email:
            return

        msg = self.parse_msg_images(msg)

        res = (
            boto3.Session(
                aws_access_key_id=self.key_id,
                aws_secret_access_key=self.access_key,
                region_name=self.region,
            )
            .client("ses")
            .send_raw_email(
                Source=sender,
                Destinations=[recipient.email],
                RawMessage={"Data": msg.as_string()},
            )
        )
        self.validate_send(res)



from typing import Optional, Union

import requests

from django.conf import Settings
from rest_framework.status import is_success

from django_project_base.notifications.base.channels.channel import Recipient
from django_project_base.notifications.base.channels.integrations.provider_integration import ProviderIntegration
from django_project_base.notifications.base.phone_number_parser import PhoneNumberParser
from django_project_base.notifications.models import DeliveryReport, DjangoProjectBaseNotification


class NexmoSMS(ProviderIntegration):
    api_key: str
    api_secret: str

    def __init__(self) -> None:
        super().__init__(settings=object())

    def ensure_credentials(self, settings: Optional[Settings] = None):
        if settings and getattr(settings, "TESTING", False):
            return
        self.api_key = getattr(settings, "NOTIFICATIONS_NEXMO_API_KEY", None)
        self.api_secret = getattr(settings, "NOTIFICATIONS_NEXMO_API_SECRET", None)
        self.settings = settings
        assert self.api_key, "NOTIFICATIONS_NEXMO_API_KEY required"
        assert self.api_secret, "NOTIFICATIONS_NEXMO_API_SECRET required"

    def _get_sms_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return super()._get_sms_message(notification)

    def validate_send(self, response: object):
        assert response
        assert is_success(response.status_code)

    def client_send(self, sender: str, recipient: Recipient, msg: str, dlr_id: str):
        # TODO: SLOVENIA????????
        rec = PhoneNumberParser.ensure_country_code_slovenia([recipient.phone_number])
        if not rec:
            return

        if not rec[0]:
            return

        params: dict = {
            "api_key": self.api_key,
            "api_secret": self.api_secret,
            "from": sender,
            "to": rec[0],
            "text": msg,
        }
        response = requests.get(
            "https://rest.nexmo.com/sms/json",
            params=params,
            verify=False,
            timeout=4,
        )
        self.validate_send(response)

    def get_message(self, notification: DjangoProjectBaseNotification) -> Union[dict, str]:
        return self._get_sms_message(notification)

    def parse_delivery_report(self, dlr: DeliveryReport):
        pass

    @property
    def delivery_report_username_setting_name(self) -> str:
        return "nexmo-sms-dlr-user"

    @property
    def delivery_report_password_setting_name(self) -> str:
        return "nexmo-sms-dlr-password"

    def ensure_dlr_user(self, project_slug: str):
        pass

    def enqueue_dlr_request(self, pk: str):
        pass



# Generated by Django 4.2.4 on 2023-08-31 10:50

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0003_alter_djangoprojectbasemessage_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="djangoprojectbasemessage",
            options={"verbose_name": "Message"},
        ),
    ]



import datetime
import uuid

from random import randrange

import django
import swapper

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group, Permission
from django.core.cache import cache
from django.db import transaction
from django.db.models import ForeignKey, Model, QuerySet
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.crypto import get_random_string
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from drf_spectacular.utils import extend_schema, extend_schema_view, OpenApiParameter, OpenApiResponse
from dynamicforms import fields
from dynamicforms.action import TableAction, TablePosition
from dynamicforms.mixins import DisplayMode
from dynamicforms.serializers import ModelSerializer
from dynamicforms.template_render.layout import Column, Layout, Row
from dynamicforms.template_render.responsive_table_layout import ResponsiveTableLayout, ResponsiveTableLayouts
from dynamicforms.viewsets import ModelViewSet
from natural.date import compress
from rest_framework import exceptions, filters, status
from rest_framework.decorators import action
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.fields import IntegerField
from rest_framework.permissions import IsAuthenticated
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import Serializer
from rest_registration.exceptions import UserNotFound

from django_project_base.account.constants import MERGE_USERS_QS_CK
from django_project_base.account.middleware import ProjectNotSelectedError
from django_project_base.account.rest.project_profiles_utils import get_project_members
from django_project_base.base.event import UserRegisteredEvent
from django_project_base.base.permissions import IsProjectOwner
from django_project_base.constants import NOTIFY_NEW_USER_SETTING_NAME
from django_project_base.notifications.email_notification import (
    EMailNotification,
    SystemEMailNotificationWithListOfEmails,
)
from django_project_base.notifications.models import DjangoProjectBaseMessage
from django_project_base.permissions import BasePermissions
from django_project_base.rest.project import ProjectSerializer, ProjectViewSet
from django_project_base.settings import DELETE_PROFILE_TIMEDELTA, USER_CACHE_KEY
from django_project_base.utils import get_pk_name

search_fields = ["username", "email", "first_name", "last_name"]


class ProfilePermissionsField(fields.ManyRelatedField):
    @staticmethod
    def to_dict(permission: Permission) -> dict:
        return {
            get_pk_name(Permission): permission.pk,
            "codename": permission.codename,
            "name": permission.name,
        }

    def to_representation(self, value, row_data=None):
        res = []
        if row_data and row_data.pk:
            if implicit_permissions := getattr(settings, "DJANGO_PROJECT_BASE_IMPLICIT_PERMISSIONS_FUNCTION", None):
                try:
                    res.extend(import_string(implicit_permissions)(row_data, self.context["request"].selected_project))
                except NotImplementedError:
                    pass

            res.extend([ProfilePermissionsField.to_dict(p) for p in row_data.user_permissions.all()])
        return res


class ProfileGroupsField(fields.ManyRelatedField):
    def to_representation(self, value, row_data=None):
        if row_data and row_data.pk:
            return [
                {
                    get_pk_name(Group): g.pk,
                    "permissions": [ProfilePermissionsField.to_dict(p) for p in g.permissions.all()],
                    "name": g.name,
                }
                for g in row_data.groups.all()
            ]
        return []


class ProfileSerializer(ModelSerializer):
    template_context = dict(url_reverse="profile-base-project")

    form_titles = {
        "table": "User profiles",
        "new": "New user",
        "edit": "Edit user",
    }

    id = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)
    username = fields.AutoGeneratedField(display_table=DisplayMode.HIDDEN)
    first_name = fields.AutoGeneratedField(display_table=DisplayMode.HIDDEN)
    last_name = fields.AutoGeneratedField(display_table=DisplayMode.HIDDEN)
    bio = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    theme = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    password = fields.AutoGeneratedField(
        password_field=True, display=DisplayMode.HIDDEN, write_only=True, required=False
    )
    last_login = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    date_joined = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    is_active = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    is_superuser = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    is_staff = fields.AutoGeneratedField(display=DisplayMode.HIDDEN)  # read_only=True,
    avatar = fields.AutoGeneratedField(display_table=DisplayMode.HIDDEN)
    reverse_full_name_order = fields.AutoGeneratedField(display_table=DisplayMode.HIDDEN)

    full_name = fields.SerializerMethodField(label=_("Full name"), display_form=DisplayMode.HIDDEN)
    is_impersonated = fields.SerializerMethodField(display=DisplayMode.HIDDEN)
    password_invalid = fields.BooleanField(display_form=DisplayMode.HIDDEN, display_table=DisplayMode.HIDDEN)

    delete_at = fields.DateTimeField(read_only=True, display=DisplayMode.HIDDEN)
    permissions = ProfilePermissionsField(
        source="user_permissions",
        child_relation=fields.PrimaryKeyRelatedField(
            help_text=_("Specific permissions for this user"), queryset=Permission.objects.all(), required=False
        ),
        help_text=_("Specific permissions for this user"),
        required=False,
        allow_null=False,
        read_only=True,
        display=DisplayMode.SUPPRESS,
    )

    groups = ProfileGroupsField(
        child_relation=fields.PrimaryKeyRelatedField(
            help_text=_(
                "The groups this user belongs to. A user will get all permissions granted to each of their groups."
            ),
            queryset=Group.objects.all(),
            required=False,
        ),
        help_text=_(
            "The groups this user belongs to. A user will get all permissions granted to each of their groups."
        ),
        required=False,
        allow_null=False,
        read_only=True,
        display=DisplayMode.SUPPRESS,
    )

    def __init__(self, *args, is_filter: bool = False, **kwds):
        super().__init__(*args, is_filter=is_filter, **kwds)
        request = self._context["request"]

        if not request.user.is_superuser:
            self.fields.pop("is_staff", None)
            self.fields.pop("is_superuser", None)

        if (
            self.instance
            and not isinstance(self.instance, (list, QuerySet, dict))
            and self.instance.pk != request.user.pk
        ):
            # only show this field to the user for their account. admins don't see this field
            self.fields.pop("reverse_full_name_order", None)

        is_record_view = len(args) == 1 and isinstance(args[0], swapper.load_model("django_project_base", "Profile"))
        if not is_record_view:
            # for table view, we remove permissions and groups to improve performance
            self.fields.pop("permissions", None)
            self.fields.pop("groups", None)

        if str(request.query_params.get("remove-merge-users", "false")) in tuple(
            map(str, fields.BooleanField.TRUE_VALUES)
        ) and (request.user.is_superuser or request.user.is_staff):
            self.actions.actions.append(
                TableAction(
                    TablePosition.ROW_END,
                    label=_("Merge"),
                    title=_("Merge"),
                    name="add-to-merge",
                    icon="git-merge-outline",
                ),
            )

        if request.user.is_superuser or request.user.is_staff:
            self.actions.actions.append(
                TableAction(
                    TablePosition.HEADER,
                    label=_("Export"),
                    title=_("Export"),
                    name="export",
                    icon="download-outline",
                )
            )

    def get_full_name(self, obj):
        # we have made it so that str(UserProfile) returns full_name, but possibly decorated with status (Klubis)
        return str(obj)

    def get_is_impersonated(self, obj):
        try:
            request = self.context["request"]
            session = request.session
            return bool(session.get("hijack_history", [])) and obj.id == request.user.id
        except:
            pass
        return False

    class Meta:
        model = swapper.load_model("django_project_base", "Profile")
        exclude = ("user_permissions",)
        layout = Layout(
            Row(Column("username"), Column("password")),
            Row(Column("first_name"), Column("last_name")),
            # Row("reverse_full_name_order"),
            Row("email"),
            Row("phone_number"),
            Row("language"),
            Row("avatar"),
            columns=2,
            size="large",
        )
        responsive_columns = ResponsiveTableLayouts(
            auto_generate_single_row_layout=True,
            auto_generate_single_column_layout=False,
            layouts=[
                ResponsiveTableLayout("full_name", "email", "#actions-row_end", auto_add_non_listed_columns=False),
                ResponsiveTableLayout("full_name", "#actions-row_end", auto_add_non_listed_columns=False),
                ResponsiveTableLayout(["full_name", "#actions-row_end"], auto_add_non_listed_columns=False),
            ],
        )


class ProfileRegisterSerializer(ProfileSerializer):
    password = fields.CharField(label=_("Password"), password_field=True)
    password_repeat = fields.CharField(label=_("Repeat Password"), password_field=True)

    class Meta(ProfileSerializer.Meta):
        layout = Layout(
            Row(Column("username")),
            Row(Column("first_name"), Column("last_name")),
            Row("password"),
            Row("password_repeat"),
            Row("email"),
            Row("phone_number"),
            Row("language"),
            columns=2,
            size="large",
        )
        exclude = ProfileSerializer.Meta.exclude + ("avatar",)

    def validate(self, attrs):
        super().validate(attrs)
        errors = {}
        password_repeat = attrs.pop("password_repeat")
        if not attrs["password"]:
            errors["password"] = _("Password is required")
        if not attrs["password"] == password_repeat:
            errors["password_repeat"] = _("Passwords do not match")

        if not attrs["email"]:
            errors["email"] = _("Email is required")

        if errors:
            raise ValidationError(errors)
        return attrs


class MergeUserRequest(Serializer):
    user = IntegerField(min_value=1, required=True, allow_null=False)


class ProfileViewPermissions(BasePermissions):
    """
    Allows users to have full permissions on get/post (retrieving and adding new users).
    Other methods require authentication.
    """

    def has_permission(self, request, view):
        if request.method == "POST":
            return True
        if request.method == "GET" and request.path.split("/")[-1].split(".")[0] == "new":
            return True
        return super().has_permission(request, view)


@extend_schema_view(
    create=extend_schema(exclude=True),
    update=extend_schema(exclude=True),
)
class ProfileViewSet(ModelViewSet):
    filter_backends = [filters.SearchFilter]
    search_fields = search_fields
    permission_classes = (ProfileViewPermissions,)
    pagination_class = ModelViewSet.generate_paged_loader(30, ["un_sort", "id"])

    def get_permissions(self):
        if self.action == "register_account":
            return []
        elif self.action == "get_current_profile":
            return [IsAuthenticated()]
        elif self.action == "merge_accounts":
            return [IsAuthenticated()]
        elif self.action == "merge":
            return [IsAuthenticated(), IsProjectOwner()]
        elif self.action == "reset_user_data":
            return [IsAuthenticated()]
        else:
            return super().get_permissions()

    def get_queryset(self):
        return get_project_members(self.request)

    def get_serializer_class(self):
        if self.request.query_params.get("select", "") == "1":

            class SearchProfileSerializer(ProfileSerializer):
                class Meta(ProfileSerializer.Meta):
                    fields = search_fields + [get_pk_name(get_user_model()), "full_name"]
                    exclude = None

            return SearchProfileSerializer

        return ProfileSerializer

    def filter_queryset(self, queryset, query_params=None):
        if filter_func := getattr(settings, "DJANGO_PROJECT_BASE_FILTER_PROFILE_QUERYSET_FIELD_FUNCTION", None):
            try:
                queryset = import_string(filter_func)(self, queryset, None, query_params)
            except NotImplementedError:
                pass
        return super().filter_queryset(queryset, query_params)

    def filter_queryset_field(self, queryset, field, value):
        if field == "full_name":
            return queryset.filter(un__icontains=value)
        if filter_func := getattr(settings, "DJANGO_PROJECT_BASE_FILTER_PROFILE_QUERYSET_FIELD_FUNCTION", None):
            try:
                return import_string(filter_func)(self, queryset, field, value)
            except NotImplementedError:
                pass

        return super().filter_queryset_field(queryset, field, value)

    @extend_schema(
        parameters=[
            OpenApiParameter(
                "search", description="Search users by all of those fields: username, email, first_name, last_name"
            )
        ],
        description="Get list of users",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK", response=get_serializer_class),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    @extend_schema(
        description="Default parameters for user registration",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @action(
        methods=["GET"],
        detail=False,
        url_path="register",
        url_name="profile-register",
    )
    def register_account(self, request: Request, **kwargs):
        register_flow_identifier = str(uuid.uuid4())
        response = Response(ProfileRegisterSerializer(None, context=self.get_serializer_context()).data)
        response.set_cookie(
            "register-flow",
            register_flow_identifier,
            max_age=settings.CONFIRMATION_CODE_TIMEOUT,
            httponly=True,
            samesite="Strict",
        )
        cache.set(register_flow_identifier, get_random_string(length=6), timeout=settings.CONFIRMATION_CODE_TIMEOUT)
        return response

    @extend_schema(
        description="Registering new account",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @register_account.mapping.post
    @transaction.atomic
    def create_new_account(self, request: Request, **kwargs):
        # set default values
        request.data["date_joined"] = datetime.datetime.now()
        request.data["is_active"] = False

        # call serializer to do the data processing drf way - hijack
        serializer = ProfileRegisterSerializer(
            None, context=self.get_serializer_context(), data=request.data, many=False
        )
        serializer.is_valid(raise_exception=True)

        if get_user_model().objects.filter(email=serializer.validated_data["email"]).exists():
            raise ValidationError({"email": _("User with this email already exists.")})

        user = serializer.save()
        user.set_password(request.data["password"])
        user.save()
        UserRegisteredEvent(user=user).trigger(payload=request)
        return Response(serializer.validated_data)

    @extend_schema(
        description="Get user profile by id",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK", response=get_serializer_class),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    def retrieve(self, request, *args, **kwargs):
        return super().retrieve(request, *args, **kwargs)

    @extend_schema(
        description="Update profile data (partially)",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK", response=get_serializer_class),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    def partial_update(self, request, *args, **kwargs):
        return super().partial_update(request, *args, **kwargs)

    @extend_schema(
        description="Get user profile of calling user.",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @action(
        methods=["GET"],
        detail=False,
        url_path="current",
        url_name="profile-current",
    )
    def get_current_profile(self, request: Request, **kwargs) -> Response:
        user: Model = request.user
        serializer = self.get_serializer(user)
        response_data: dict = serializer.data
        cache_key = f"user-current-project-{user.pk}"
        try:
            response_data["default_project"] = ProjectSerializer(request.selected_project).data
            cache.set(cache_key, request.selected_project.pk, timeout=None)
        except ProjectNotSelectedError:
            q = ProjectViewSet._get_queryset_for_request(request)
            # todo this might be a problem if the cache is cleared. then selected project might change for some users as
            #   previously selected project would be forgotten. might have to move this to a table?
            previously_selected_project_pk = cache.get(cache_key)
            if project_object := q.filter(pk=previously_selected_project_pk).first():
                response_data["default_project"] = ProjectSerializer(project_object).data
            elif project_object := q.first():
                response_data["default_project"] = ProjectSerializer(project_object).data
            else:
                response_data["default_project"] = None

        return Response(response_data)

    @extend_schema(
        description="Marks profile of calling user for deletion in future. Future date is determined " "by settings",
        responses={
            status.HTTP_200_OK: OpenApiResponse(description="OK"),
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @get_current_profile.mapping.put
    def update_current_profile(self, request: Request, **kwargs) -> Response:
        user: Model = request.user
        new_email = request.data.pop("email", None)
        serializer = self.get_serializer(user, data=request.data, many=False)
        serializer.is_valid(raise_exception=True)
        serializer.save()
        email_changed = new_email and user.email != new_email
        response = Response(serializer.data)
        if email_changed:
            code = randrange(100001, 999999)
            response.set_cookie("verify-email", user.pk, samesite="Lax")
            request.session[f"email-changed-{code}-{user.pk}"] = new_email
            SystemEMailNotificationWithListOfEmails(
                message=DjangoProjectBaseMessage(
                    subject=f"{_('Email change for account on')} {request.META['HTTP_HOST']}",
                    body=f"{_('You requested an email change for your account at')} {request.META['HTTP_HOST']}. "
                    f"\n\n{_('Your verification code is')}: "
                    f"{code} \n\n {_('Code is valid for')} {compress(settings.CONFIRMATION_CODE_TIMEOUT)}.\n",
                    footer="",
                    content_type=DjangoProjectBaseMessage.PLAIN_TEXT,
                ),
                recipients=[new_email],
                project=self.request.selected_project.slug,
                user=user.pk,
            ).send()
        return response

    @extend_schema(exclude=True)
    @action(
        methods=["POST"],
        detail=False,
        url_path="confirm-new-email",
        url_name="confirm-new-email",
    )
    @transaction.atomic()
    def confirm_new_email(self, request: Request, **kwargs) -> Response:
        user = request.user
        if not request.data.get("code"):
            raise ValidationError(dict(code=[_("Code required")]))
        key = f"email-changed-{request.data['code']}-{user.pk}"
        new_email = request.session.get(key)
        if email := new_email:
            user.email = email
            user.save(update_fields=["email"])
            request.session.pop(key, None)
            response = Response()
            response.delete_cookie("verify-email")
            return response
        raise ValidationError(dict(code=[_("Invalid code")]))

    @extend_schema(
        description="Marks profile of calling user for deletion in future. Future date is determined " "by settings",
        responses={
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
            status.HTTP_403_FORBIDDEN: OpenApiResponse(description="Not allowed"),
        },
    )
    @get_current_profile.mapping.delete
    def mark_current_profile_delete(self, request: Request, **kwargs) -> Response:
        user: Model = getattr(request, "user", None)
        if not user:
            raise exceptions.AuthenticationFailed
        # user.is_active = False // user must still be able to login
        profile_obj = getattr(user, swapper.load_model("django_project_base", "Profile")._meta.model_name)
        profile_obj.delete_at = timezone.now() + datetime.timedelta(days=DELETE_PROFILE_TIMEDELTA)

        profile_obj.save(update_fields=["delete_at"])
        # user.save(update_fields=["is_active"])
        cache.delete(USER_CACHE_KEY.format(id=user.id))
        request.session.flush()
        return Response(status=status.HTTP_204_NO_CONTENT)

    @extend_schema(
        description="Immediately removes user from database",
        responses={
            status.HTTP_204_NO_CONTENT: OpenApiResponse(description="No content"),
        },
    )
    def destroy(self, request, *args, **kwargs):
        if self.request.user.is_superuser or self.request.user.is_staff:
            return super().destroy(request, *args, **kwargs)
        raise exceptions.PermissionDenied

    @extend_schema(exclude=True)
    @action(methods=["POST"], detail=False, url_path="merge-accounts", url_name="merge-accounts")
    def merge_accounts(self, request, *args, **kwargs):
        from rest_registration.settings import registration_settings

        serializer = registration_settings.LOGIN_SERIALIZER_CLASS(data=request.data)
        serializer.is_valid(raise_exception=True)
        auth_user = self.request.user
        auth_user_is_main = str(self.request.data.get("account", "false")) in fields.BooleanField.TRUE_VALUES
        try:
            from django_project_base.account.service.merge_users_service import MergeUsersService

            user = registration_settings.LOGIN_AUTHENTICATOR(serializer.validated_data, serializer=serializer)
            MergeUserGroup = swapper.load_model("django_project_base", "MergeUserGroup")

            group, created = MergeUserGroup.objects.get_or_create(
                users=f"{auth_user.pk},{user.pk}", created_by=self.request.user.pk
            )
            MergeUsersService().handle(user=auth_user if auth_user_is_main else user, group=group)
            if not auth_user_is_main:
                # logout current user and redirect to login
                request.session.flush()
                return Response(status=status.HTTP_401_UNAUTHORIZED)
        except UserNotFound:
            raise UserNotFound
        except Exception:
            raise APIException

    @extend_schema(exclude=True)
    @action(
        methods=["POST"],
        detail=False,
        url_path="merge",
        url_name="merge",
    )
    def merge(self, request: Request, **kwargs) -> Response:
        ser = MergeUserRequest(data=request.data)
        ser.is_valid(raise_exception=True)
        ck = MERGE_USERS_QS_CK % self.request.user.pk
        ck_val = cache.get(ck, [])
        ck_val.append(ser.validated_data["user"])
        cache.set(ck, list(set(ck_val)), timeout=None)
        return Response(status=status.HTTP_201_CREATED)

    @extend_schema(exclude=True)
    @action(
        methods=["POST"],
        detail=False,
        url_path="reset-user-data",
        url_name="reset-user-data",
    )
    @transaction.atomic()
    def reset_user_data(self, request: Request, **kwargs) -> Response:
        profile_model = swapper.load_model("django_project_base", "Profile")
        profile_obj = getattr(request.user, profile_model._meta.model_name)
        if request.data.get("reset"):
            base_user_models = (get_user_model(), profile_model)
            for mdl in django.apps.apps.get_models(include_auto_created=True, include_swapped=True):
                if mdl not in base_user_models and not mdl._meta.abstract and not mdl._meta.swapped:
                    for fld in [
                        f
                        for f in mdl._meta.fields
                        if isinstance(f, ForeignKey) and (f.related_model in base_user_models)
                    ]:
                        mdl.objects.filter(**{fld.attname: fld.to_python(profile_obj.pk)}).delete()
        profile_obj.delete_at = None
        profile_obj.save(update_fields=["delete_at"])
        return Response(status=status.HTTP_204_NO_CONTENT)

    @transaction.atomic()
    def create(self, request, *args, **kwargs):
        response = super().create(request, *args, **kwargs)
        project = swapper.load_model("django_project_base", "Project").objects.get(
            slug=getattr(self.request, settings.DJANGO_PROJECT_BASE_BASE_REQUEST_URL_VARIABLES["project"]["value_name"])
        )
        if (
            sett := swapper.load_model("django_project_base", "ProjectSettings")
            .objects.filter(name=NOTIFY_NEW_USER_SETTING_NAME, project=project)
            .first()
        ) and sett.python_value:
            recipients = [response.data[get_pk_name(get_user_model())]]
            EMailNotification(
                message=DjangoProjectBaseMessage(
                    subject=_("Your account was created for you"),
                    body=render_to_string(
                        "account_created.html",
                        {
                            "username": f"{response.data['username']}/{response.data['email']}",
                        },
                    ),
                    footer="",
                    content_type=DjangoProjectBaseMessage.HTML,
                ),
                raw_recipents=recipients,
                project=project.slug,
                recipients=recipients,
                delay=int(datetime.datetime.now().timestamp()),
                user=self.request.user.pk,
            ).send()

        return response


class ProjectsProfileSearchViewSet(ProfileViewSet):
    def get_queryset(self):
        projects = set(map(lambda pm: pm.project, self.request.user.projects.all()))
        first_project = next(iter(projects), None)
        if not first_project:
            return swapper.load_model("django_project_base", "Profile").objects.none()
        qs = get_project_members(self.request, project=first_project)
        projects = projects - {first_project}
        for project in projects:
            qs = qs | get_project_members(self.request, project=project)
        return qs.distinct()



import os

from celery import bootsteps, Celery
from click import Option
from django.apps import apps
from django.utils.crypto import get_random_string
from kombu import Exchange, Queue
from kombu.entity import TRANSIENT_DELIVERY_MODE

from django_project_base.celery.settings import NOTIFICATIONS_QUEUE_VISIBILITY_TIMEOUT
from django_project_base.constants import NOTIFICATION_QUEUE_NAME

REDIS_LOCATION = ""
CELERY_REDIS_DB = "3"
redis_location_celery = ""

if os.path.exists("/var/run/redis/redis-server.sock"):
    REDIS_LOCATION = "/var/run/redis/redis-server.sock"

if REDIS_LOCATION:
    redis_location_celery = f"redis+socket://{REDIS_LOCATION}"
    CELERY_BROKER_URL = f"{redis_location_celery}{'?virtual_host=' + CELERY_REDIS_DB}"


class CelerySettings:
    accept_content = ["application/json", "application/x-python-serialize", "pickle"]
    TASK_SERIALIZER = "json"
    RESULT_SERIALIZER = "json"
    TIMEZONE = "UTC"
    broker_url = f"{redis_location_celery}{'?virtual_host=' + CELERY_REDIS_DB}" if REDIS_LOCATION else ""
    SECRET_KEY = get_random_string(length=64)
    INSTALLED_APPS = (
        "django_project_base",
        "django_project_base.licensing",
        "django_project_base.notifications",
        "django.contrib.contenttypes",
        "django.contrib.auth",
    )
    TESTING = False


os.environ.setdefault("DJANGO_SETTINGS_MODULE", "django_project_base.celery.settings")
apps.populate(CelerySettings().INSTALLED_APPS)

app = Celery(
    "django_project_base",
    config_source=CelerySettings,
    include=[
        "django_project_base.celery.background_tasks.notification_tasks",
    ],
)

app.conf.task_queues = [
    Queue(
        NOTIFICATION_QUEUE_NAME,
        Exchange("transient", delivery_mode=TRANSIENT_DELIVERY_MODE),
        routing_key=NOTIFICATION_QUEUE_NAME,
        durable=False,
    ),
]

app.conf.task_ignore_result = True
app.conf.worker_send_task_events = False
app.conf.broker_transport_options = {"visibility_timeout": NOTIFICATIONS_QUEUE_VISIBILITY_TIMEOUT}
setting_option = Option(("--settings",), is_flag=False, help="Django settings file path", default="")
app.user_options["worker"].add(setting_option)


class CeleryBootstep(bootsteps.Step):
    def __init__(self, parent, **options):
        super().__init__(parent, **options)
        app.conf.setdefault("django-settings-module", options.get("settings", ""))


app.steps["worker"].add(CeleryBootstep)

# RUN WORKER AS
# celery -A django_project_base.celery.celery worker -l INFO -Q notification --concurrency=1



# Generated by Django 4.2.4 on 2023-10-26 05:55
from gettext import gettext

import swapper
from django.db import migrations

from django_project_base.constants import USE_EMAIL_IF_RECIPIENT_HAS_NO_PHONE_NUMBER


def forwards_func(apps, schema_editor):
    project_sett = swapper.load_model("django_project_base", "ProjectSettings")
    for project in swapper.load_model("django_project_base", "Project").objects.all():
        project_sett.objects.get_or_create(
            project=project,
            name=USE_EMAIL_IF_RECIPIENT_HAS_NO_PHONE_NUMBER,
            defaults=dict(
                description=gettext("Send notification via EMail if user has no phone number"),
                value=False,
                value_type="bool",
            ),
        )


def reverse_func(apps, schema_editor):
    pass


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0006_djangoprojectbasenotification_send_notification_sms"),
    ]

    operations = [
        migrations.RunPython(forwards_func, reverse_func),
    ]



import datetime
import uuid

from typing import List, Optional, Type

from django.conf import settings
from django.core.cache import cache
from rest_framework.exceptions import PermissionDenied

from django_project_base.notifications.base.channels.channel import Channel
from django_project_base.notifications.base.channels.mail_channel import MailChannel
from django_project_base.notifications.base.enums import NotificationLevel, NotificationType
from django_project_base.notifications.base.notification import Notification
from django_project_base.notifications.models import DjangoProjectBaseMessage, DjangoProjectBaseNotification


class EMailNotification(Notification):
    def __init__(
        self,
        message: DjangoProjectBaseMessage,
        raw_recipents,
        project,
        persist: bool = True,
        level: Optional[NotificationLevel] = None,
        locale: Optional[str] = None,
        delay: Optional[int] = None,
        type: Optional[NotificationType] = None,
        recipients=None,
        **kwargs,
    ) -> None:
        super().__init__(
            message=message,
            raw_recipents=raw_recipents,
            project=project,
            persist=persist,
            level=level,
            locale=locale,
            delay=delay,
            type=type,
            recipients=recipients,
            **kwargs,
        )

    @property
    def via_channels(self) -> List[Type[Channel]]:
        return [MailChannel]


class EMailNotificationWithListOfEmails(EMailNotification):
    def __init__(self, message: DjangoProjectBaseMessage, recipients, project, user, **kwargs) -> None:
        super().__init__(
            raw_recipents=recipients,
            message=message,
            project=project,
            recipients=recipients,
            persist=True,
            delay=int(datetime.datetime.now().timestamp()),
            user=user,
            **kwargs,
        )

    def send(self) -> DjangoProjectBaseNotification:
        from django_project_base.notifications.base.channels.mail_channel import MailChannel

        notification: DjangoProjectBaseNotification = DjangoProjectBaseNotification(
            delayed_to=self.delay,
            required_channels=f"{MailChannel.name}",
            message=self.message,
            content_entity_context=str(self.content_entity_context)
            if self.content_entity_context
            else str(uuid.uuid4()),
            recipients=",".join(map(str, self._recipients)) if self._recipients else None,
            recipients_original_payload=self._raw_recipents,
            project_slug=self._project,
        )

        notification = self._ensure_channels([MailChannel.name], notification, settings=settings)

        if self.handle_similar_notifications(notification=notification):
            return notification
        if not self.message.pk or not DjangoProjectBaseMessage.objects.filter(pk=self.message.pk).exists():
            self.message.save()
        notification.created_at = int(datetime.datetime.now().timestamp())
        notification.save()
        uuid_val = str(uuid.uuid4())
        notification.email_list = [
            dict(
                id=uuid_val,
                email=u,
                phone_number="",
            )
            for u in self._recipients
        ]
        self.enqueue_notification(notification, self._extra_data)
        return notification


class SystemEMailNotification(EMailNotification):
    system_mail_throttling_ck = "system-mail-throttling-ck"
    allowed_number_of_system_requests_per_minute = 5

    def __init__(
        self,
        message: DjangoProjectBaseMessage,
        recipients,
        **kwargs,
    ) -> None:
        super().__init__(
            message,
            raw_recipents=recipients,
            project=None,
            persist=True,
            level=None,
            locale=None,
            delay=int(datetime.datetime.now().timestamp()),
            type=None,
            recipients=recipients,
            is_system_notification=True,
            **kwargs,
        )

    def _get_system_email_cache_key(self) -> str:
        return f"{self.system_mail_throttling_ck}-{datetime.datetime.now().minute}"

    def _get_system_email_cache_value(self) -> List[float]:
        if ck_val := cache.get(self._get_system_email_cache_key()):
            return ck_val
        return []

    def _check_request_limit(self):
        if len(self._get_system_email_cache_value()) > self.allowed_number_of_system_requests_per_minute:
            raise PermissionDenied

    def _register_system_email(self):
        if ck_val := self._get_system_email_cache_value():
            if len(ck_val) > self.allowed_number_of_system_requests_per_minute + 1:
                return
            ck_val.append(datetime.datetime.now().timestamp())
            cache.set(self._get_system_email_cache_key(), ck_val, timeout=70)
            return
        cache.set(self._get_system_email_cache_key(), [datetime.datetime.now().timestamp()], timeout=70)

    def send(self) -> DjangoProjectBaseNotification:
        # throttling for system messages
        self._check_request_limit()
        self._register_system_email()
        return super().send()


class SystemEMailNotificationWithListOfEmails(EMailNotificationWithListOfEmails):
    def __init__(
        self,
        message: DjangoProjectBaseMessage,
        recipients,
        **kwargs,
    ) -> None:
        super().__init__(
            message,
            recipients,
            project=kwargs.pop("project", None),
            user=kwargs.pop("user", None),
            level=kwargs.pop("level", None),
            locale=kwargs.pop("locale", None),
            type=kwargs.pop("type", None),
            is_system_notification=True,
            **kwargs,
        )



# Generated by Django 4.2.4 on 2023-08-18 12:58

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="LicenseAccessUse",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "date",
                    models.DateTimeField(default=django.utils.timezone.now, verbose_name="Date"),
                ),
                (
                    "type",
                    models.IntegerField(
                        choices=[(4, "Used access"), (10, "Used access by admin")],
                        default=4,
                        verbose_name="Type",
                    ),
                ),
                ("user_id", models.CharField(db_index=True, max_length=1024)),
                ("content_type_object_id", models.CharField(max_length=1024)),
                (
                    "comment",
                    models.CharField(max_length=255, null=True, verbose_name="Comment"),
                ),
                (
                    "content_type",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.DO_NOTHING,
                        to="contenttypes.contenttype",
                    ),
                ),
            ],
        ),
    ]



# Generated by Django 4.2.4 on 2023-10-26 06:52

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("notifications", "0007_auto_20231026_0555"),
    ]

    operations = [
        migrations.AddField(
            model_name="deliveryreport",
            name="auxiliary_notification",
            field=models.UUIDField(null=True, verbose_name="Auxiliary notification"),
        ),
    ]



from django.core.cache import caches


class CacheCounter:
    cache = None
    key = None
    timeout = None

    def __init__(self, key, cache_name="default", timeout=-1):
        self.cache = caches[cache_name]
        self.set_timeout(timeout)
        self.key = key

    def get_default_timeout(self):
        return self.cache.default_timeout

    def set_timeout(self, timeout):
        if timeout == -1:
            timeout = self.get_default_timeout()
        self.timeout = timeout

    def update_timeout(self):
        if self.timeout is None:
            persist = getattr(self.cache, "persist", None)
            if callable(persist):
                persist(self.key)
                return
        self.cache.touch(self.key, self.timeout)

    def incr(self, step=1, start=0):
        while True:
            try:
                self.cache.add(self.key, start)
                ret = self.cache.incr(self.key, step)
                self.update_timeout()
                return ret
            except ValueError:  # pragma: no cover
                # This happens if another thread just deleted the cache entry after our .add and before our .incr
                pass



import uuid

from typing import List

from django.conf import Settings
from django.core.exceptions import ValidationError
from django.core.validators import validate_email

from django_project_base.notifications.base.channels.channel import Channel, Recipient
from django_project_base.notifications.base.enums import ChannelIdentifier
from django_project_base.notifications.models import DjangoProjectBaseNotification


class MailChannel(Channel):
    id = ChannelIdentifier.MAIL.value

    name = "EMail"

    notification_price = 0.0002  # TODO get from settings

    provider_setting_name = "NOTIFICATIONS_EMAIL_PROVIDER"

    def send(self, notification: DjangoProjectBaseNotification, extra_data, settings: Settings, **kwargs) -> int:
        if getattr(settings, "TESTING", False):
            return super().send(notification=notification, extra_data=extra_data, settings=settings)
        message = self.provider.get_message(notification)
        sender = self.sender(notification)
        self.provider.client_send(
            self.sender(notification),
            Recipient(identifier=str(uuid.uuid4()), phone_number="", email=sender),
            message,
            str(uuid.uuid4()),
        )
        return super().send(notification=notification, extra_data=extra_data, settings=settings)

    def get_recipients(
        self, notification: DjangoProjectBaseNotification, unique_identifier="", phone_number_validator=None
    ):
        return list(
            set(
                super().get_recipients(
                    notification, unique_identifier="email", phone_number_validator=phone_number_validator
                )
            )
        )

    def clean_email_recipients(self, recipients: List[Recipient]) -> List[Recipient]:
        valid = []
        for email in self.clean_recipients(recipients):
            try:
                validate_email(email.email)
                valid.append(email)
            except ValidationError:
                pass
        return valid



from abc import ABC, abstractmethod
from typing import Optional

from django.conf import settings
from django.core.cache import cache, caches


class CacheQueue(ABC):
    key = None
    cache = None
    timeout = None
    django_cache = None

    def __init__(self, key, cache_name, timeout):
        self.cache_name = cache_name
        self.django_cache = caches[self.cache_name]
        self.set_cache()
        self.key = key
        self.set_timeout(timeout)

    @abstractmethod
    def set_cache(self):
        """Set cache client"""

    @abstractmethod
    def rpush(self, *values):
        """Add data to end of queue"""

    @abstractmethod
    def lpush(self, *values):
        """Add data to start of queue"""

    @abstractmethod
    def lpop(self, count: Optional[int] = None):
        """Get and remove data from start of queue"""

    @abstractmethod
    def rpop(self, count: Optional[int] = None):
        """Get and remove data from end of queue"""

    @abstractmethod
    def lrange(self, count=None):
        """Get data from start of queue"""

    @abstractmethod
    def ltrim(self, count=None):
        """Remove data from start of queue"""

    def get_default_timeout(self):
        return self.django_cache.default_timeout

    def set_timeout(self, timeout):
        if timeout == -1:
            timeout = self.get_default_timeout()
        self.timeout = timeout

    @abstractmethod
    def update_timeout(self):
        """Update timeout of cached key"""

    @staticmethod
    def is_redis_cache_backend(cache_name):
        cache_key = f"redis_cache_backend_{cache_name}"
        is_redis_backend = cache.get(cache_key, None)
        if is_redis_backend is None:
            import warnings

            from django.utils.module_loading import import_string
            from django_redis.cache import RedisCache

            backend_path = settings.CACHES[cache_name]["BACKEND"]
            backend_class = import_string(backend_path)
            is_redis_backend = issubclass(backend_class, RedisCache)

            if is_redis_backend:
                try:
                    from django_redis import get_redis_connection
                    from packaging import version

                    conn = get_redis_connection(cache_name)

                    # Execute the INFO command
                    info = conn.info()

                    # Get the Redis version
                    redis_version = info.get("redis_version")

                    if version.parse(redis_version) < version.parse("6.2"):
                        # we need django_redis installed and redis server must be greater than 6.2.0
                        warnings.warn(
                            "You are using redis cache and have django-redis package installed, "
                            "but redis server version is older than 6.2.0 which is needed for redis-optimised queue. "
                            "We will be using a non-optimised queue instead."
                        )
                        return False
                except ModuleNotFoundError:
                    warnings.warn(
                        "You are using redis cache, but django-redis package is not installed. "
                        "If it were installed, we would be using redis-optimised Queue"
                    )
                    return False
            else:
                warnings.warn("Cache backend is not RedisCache. We will be using a non-optimised queue.")

            cache.set(cache_key, is_redis_backend)

        return is_redis_backend

    @staticmethod
    def get_cache_queue(key, cache_name="default", timeout=-1):
        if CacheQueue.is_redis_cache_backend(cache_name):
            from django_project_base.caching.cache_queue.cache_queue_redis import CacheQueueRedis

            return CacheQueueRedis(key, cache_name=cache_name, timeout=timeout)
        else:
            from django_project_base.caching.cache_queue.cache_queue_other import CacheQueueOther

            return CacheQueueOther(key, cache_name=cache_name, timeout=timeout)



import uuid

from typing import List, Optional, Type

from django_project_base.notifications.base.channels.channel import Channel
from django_project_base.notifications.base.enums import NotificationLevel, NotificationType
from django_project_base.notifications.base.notification import Notification
from django_project_base.notifications.models import DjangoProjectBaseMessage


class MaintenanceNotification(Notification):
    def __init__(self, delay: int, message: DjangoProjectBaseMessage, locale: Optional[str] = None) -> None:
        super().__init__(
            message=message,
            raw_recipents=[],
            project=None,
            persist=True,
            level=NotificationLevel.WARNING.value,
            locale=locale,
            delay=delay,
            type=NotificationType.MAINTENANCE.value,
            content_entity_context=str(uuid.uuid4()),
        )

    @property
    def via_channels(self) -> List[Type[Channel]]:
        return []



import json
import random

from datetime import datetime

from django.core.cache import cache
from django.shortcuts import render
from dynamicforms.struct import Struct

from django_project_base.caching.cache_queue import CacheQueue
from django_project_base.settings import PROFILER_LOG_LONG_REQUESTS_COUNT


def app_debug_view(request):
    # if not getattr(request, 'user', None) or request.user.pk not in POWER_USERS:
    if not getattr(request, "user", None):
        raise PermissionError
    return render(request, "app-debug/main.html", __get_debug_data())


def __get_debug_data():
    import time

    result_data = []
    requests = []

    for cache_ptr in range(PROFILER_LOG_LONG_REQUESTS_COUNT):
        item = cache.get("long_running_cmds_data%d" % cache_ptr)
        if item:
            requests.append(item)

    totals = {}
    min_timestamp = time.time()

    for req in requests:
        queries_executed: list = req.get("queries", []) or []

        num_of_queries: int = len(queries_executed)
        num_of_distinct_queries: int = len(
            set(map(lambda d: d["sql"], filter(lambda e: isinstance(e, dict), queries_executed)))
        )
        r = lambda: random.randint(0, 255)  # noqa: E731
        item = dict(
            r_data=dict(
                num_of_duplicate_queries=num_of_queries - num_of_distinct_queries,
                num_of_queries=num_of_queries,
                duration=req.get("duration"),
                path_info=req.get("PATH_INFO"),
                host=req.get("HTTP_HOST"),
                method=req.get("REQUEST_METHOD"),
                timestamp=datetime.utcfromtimestamp(req["timestamp"]).strftime("%Y-%m-%d %H:%M:%S")
                if req.get("timestamp")
                else None,
                query_string=req.get("QUERY_STRING"),
            ),
            db_queries=queries_executed,
            color="rgba(%d, %d, %d, 0.3)" % (r(), r(), r()),
        )
        item_data = item["r_data"]
        totals.setdefault(item_data["path_info"], Struct(count=0, time=0, path=""))
        total = totals[item_data["path_info"]]
        total.count += 1
        total.time += item_data["duration"]
        total.path = item_data["path_info"]
        min_timestamp = min(req["timestamp"], min_timestamp)
        result_data.append(item)

    result_data.sort(key=lambda f: f.get("r_data", {}).get("duration", 0) or 0, reverse=True)
    spenders = list(sorted(totals.values(), key=lambda x: x.time, reverse=True))

    totals = {}
    for interval in range(int((time.time()) - 3600) // 10, int(time.time()) // 10 + 1):
        last_hour_running_cmds_key = f"last_hour_running_cmds{interval}"
        last_hour_running_cmds_queue = CacheQueue.get_cache_queue(last_hour_running_cmds_key, timeout=3600)

        cache_ptr = last_hour_running_cmds_queue.lrange()
        for item in [json.loads(item) for item in cache_ptr]:
            totals.setdefault(item[0], Struct(count=0, wall_time=0, path="", user_time=0, sys_time=0, cpu_time=0))
            total = totals[item[0]]
            total.count += 1
            total.wall_time += item[1]
            total.user_time += item[2]
            total.sys_time += item[3]
            total.cpu_time += item[2] + item[3]
            total.path = item[0]
    for total in totals.values():
        total.wall_avg = int(total.wall_time / total.count)
        total.cpu_avg = int(total.cpu_time / total.count)
        total.core_usage = int(total.cpu_time / 3600.0) / 1000.0

    all_requests = list(sorted(totals.values(), key=lambda x: x.wall_time, reverse=True))

    return dict(
        debug_data=result_data,
        spenders=spenders,
        long_running_time=int(time.time() - min_timestamp),
        all_requests=all_requests,
    )
